/*******************************************************************************
 *
 *	force.c
 *
 *	part of the layout program.
 *
 *
 *	Written by Gerard Leurs, NIKHEF-H, Amsterdam.
 *	December 1991.
 *	email: a89@nikhef.nl
 *
 *
 *	This part determines the attracting forces between connecting nodes,
 *	and the repulsing forces between non-connecting nodes.
 *	The resulting coordinates are determined from this resulting force and
 *	the original coordinates.
 *	There is a quick check on overlap for the new coordinates of a node and
 *	the coordinates of already changed nodes.
 *
 ******************************************************************************/

# include "layout.h"


	/* get the height for the more_link_node */

extern long int get_height(node)
struct hostentity * node;
{
	struct nodeheightentity * dummy;
	long int boxheight = 100;


	dummy = nodeheight;
	while (dummy != NULL)
	{
		if (node->number == dummy->number)
			return dummy->totalheight;
		dummy = dummy->next;
	}

	fprintf(stderr,"get_height: failed for %s (returning height %ld)\n",node->hostname, boxheight);
	return boxheight;
}

	/* give the coordinates and the sizes of the node */

extern int get_characteristics(hname,x,y,w,h)
struct hostentity * hname;	/* nodename we are looking for */
long int *x, *y;
long int *w, *h;		/* x_coord, y_coord, width, heigth */
{
	struct hostentity * dummyhost1;

	dummyhost1 = host;

	while (dummyhost1 != NULL)
	{
		if (strcmp(hname->hostname,dummyhost1->hostname) == 0)
		{
			*x = dummyhost1->x_coord;	*y = dummyhost1->y_coord;
			*w = dummyhost1->width;
			*h = get_height(hname) + 5;

			return TRUE;
		}

		dummyhost1 = dummyhost1->next;
	}

	fprintf(stderr,"get_characteristics: %s not found in hoststructure\n",hname);
	*x = *y = *w = *h = NONSENSE;
	return FALSE;
}


	/* determine the effective center of the node */

extern void nearest_point (x1, x2, w1, w2, xr1, xr2)
long int x1, x2;		/* coordinates */
long int w1, w2;		/* sizes */
long int *xr1, *xr2;		/* effective resulting coordinate */
{
	/* not overlapping */ 

	if (x1 + w1 <= x2)
	{
		*xr1 = x1 + w1;		*xr2 = x2;
	}
	else if (x2 + w2 <= x1)
	{
		*xr1 = x1;		*xr2 = x2 + w2;
	}

	/* overlapping, use mid point of overlapping region */

	else if (x1 <= x2)
	{
		if (x1 + w1 < x2 + w2)
			*xr1 = *xr2 = x2 + (x1 + w1 - x2) / 2;
		else
			*xr1 = *xr2 = x2 + w2 / 2;
	}
	else /* if (x1 > x2) */
	{
		if (x1 + w1 > x2 + w2)
			*xr1 = *xr2 = x1 + (x2 + w2 - x1) / 2;
		else
			*xr1 = *xr2 = x1 + w1 / 2;
	}
}


	/* determine vector between the two interacting nodes */

extern void determine_vector(hname1, hname2, vect_x, vect_y, factor)
struct hostentity * hname1;
struct hostentity * hname2;
long int *vect_x, *vect_y;		/* x/y vector for hname1 */
float factor;				/* multiply the vector by this */
{
	long int x_one, y_one;			/* coordinates for first node */
	long int w_one, h_one;			/* sizes for first node */
	long int x_two, y_two;			/* coordinates for second node */
	long int w_two, h_two;			/* sizes for second node */
	long int x1_center, y1_center;		/* symbolcenter node1 */
	long int x2_center, y2_center;		/* symbolcenter node2 */

	/* get characteristics for the two nodes */

	if ( ! get_characteristics(hname1, &x_one, &y_one, &w_one, &h_one) )
		fprintf(stderr,"determine_vector : error in get_characteristic for %s\n",hname1->hostname);

	if ( ! get_characteristics(hname2, &x_two, &y_two, &w_two, &h_two) )
		fprintf(stderr,"determine_vector : error in get_characteristic for %s\n",hname2->hostname);

	nearest_point(x_one, x_two, w_one, w_two, &x1_center, &x2_center);
	nearest_point(y_one, y_two, h_one, h_two, &y1_center, &y2_center);

	*vect_x = (x1_center - x2_center) * factor;
	*vect_y = (y1_center - y2_center) * factor;
}


	/* determine resulting force between the two nodes
	   if linked -> attracting forces
	   if not linked -> repulsing forces */

extern void determine_force(host1, host2, force_x, force_y, factor)
struct hostentity * host1;		/* the node we are determining the total forces for */
struct hostentity * host2;		/* the other nodes in the nodesfile */
long int *force_x, *force_y;		/* force vectors resulting from interacting hosts */
float factor;				/* multiply the vector by this factor */
{
	long int vector_x, vector_y;	/* vector between centers of first and second node */
	long int vect;
	long int sq_vect_x,sq_vect_y;

#ifdef DEBUG
	fprintf(stderr,"determine_force: forces of %s on %s\n",host2->hostname,host1->hostname);
#endif

	determine_vector(host1, host2, &vector_x, &vector_y, factor);

	if (linked_or_not(host1->number, host2->number))
	{
		*force_x = -vector_x;
		*force_y = -vector_y;
	}
	else
	{
		sq_vect_y = vector_y * vector_y;
		if (vector_y < 0)
			sq_vect_y = - sq_vect_y;
		vect = (vector_x == 0) ? sq_vect_y : (vector_x + sq_vect_y / vector_x);
		*force_x = (vect == 0) ? (WINDOWwidth / 4) : ((WINDOWwidth / 2) / vect);

		sq_vect_x = vector_x * vector_x;
		if (vector_x < 0)
			sq_vect_x = - sq_vect_x;
		vect = (vector_y == 0) ? sq_vect_x : (sq_vect_x / vector_y + vector_y);
		*force_y = (vect == 0) ? (WINDOWheight / 4) : ((WINDOWheight / 2) / vect );
	}
}


	/* overlap between more_links_node and the more_links_node with two connections to
	   more_links_node */

static int check_nodes(hoststr, node, x, y)
struct hostentity * hoststr;	/* the structure with all new coordinates */
struct hostentity * node;	/* the node which we want to place */
long int x, y;			/* the coordinates for this node */
{
	struct hostentity * dummyhost1;
	long int w, h;
	long int x1, y1, w1, h1;

#ifdef DEBUG
	fprintf(stderr,"check_nodes: trying to place %s at %ld %ld\n",node->hostname,x,y);
#endif

	dummyhost1 = hoststr;
	while (dummyhost1 != NULL)
	{
		if ((strcmp(dummyhost1->hostname,node->hostname)) != 0)
		{
			w = node->width;		h = get_height(node);
			x1 = dummyhost1->x_coord;	y1 = dummyhost1->y_coord;
			w1 = dummyhost1->width;		h1 = get_height(dummyhost1);

			/* overlap in x-coordinates (width) -> check y-coordinates */

			if ( ((x <= x1) && (x + w >= x1)) || ((x1 <= x) && (x1 + w1 >= x)) )
			{
				if ( (y <= y1) && ((y + h) >= y1) )
					return FALSE;
				else if ( (y1 <= y) && ((y1 + h1) >= y) )
					return FALSE;

			}
		}

		dummyhost1 = dummyhost1->next;
	}
	return TRUE;
}


	/* overlap between already changed nodes (after forcevector determination)
	   and the node of which we have just determined the new coordinates for */

static void check_overlap_layout(newhost, node, new_x, new_y, x_change, y_change)
struct hostentity * newhost;
struct hostentity * node;
long int *new_x, *new_y;
long int x_change, y_change;
{
	struct hostentity * dummyhost1;
	int No_Overlap;
	int Found = FALSE;
	long int save_x, save_y;
	long int x, y;			/* temp. coordinates (new -> old) */
	long int x_step, step_nr;

	x_step = x_change / 3;
	if (x_step == 0)
	{
		if (x_change < 0)
			x_step = -1;
		else if (x_change >= 0)
			x_step = 1;
	}
	step_nr = 0;

	/* check overlap for all coordinates, starting at the new and ending
	   at the old ones. The first one found without overlapping other
	   already changed nodes is returned */

	x = *new_x;	y = *new_y;
	if (x_change >= 0)
	{
		while ( (! Found) && (x <= (*new_x + x_change)) && (step_nr < 3))
		{
			y += step_nr * (y_change / 3);
			step_nr ++;
			dummyhost1 = newhost;
			if (dummyhost1 == NULL)
			{
				Found = TRUE;	save_x = x;	save_y = y;
			}
			if (!Found)
			{
				No_Overlap = check_nodes(newhost, node, x, y);
				if (No_Overlap)
				{
					Found = TRUE;	save_x = x;	save_y = y;
				}
			}
			x += x_step;
		}
	}
	else
	{
		while ( (! Found) && (x <= (*new_x - x_change)) && (step_nr < 3))
		{
			y = *new_y + step_nr * (y_change / 3);
			step_nr ++;
			dummyhost1 = newhost;
			if (dummyhost1 == NULL)
			{
				Found = TRUE;	save_x = x;	save_y = y;
			}
			if (!Found)
			{
				No_Overlap = check_nodes(newhost, node, x, y);
				if (No_Overlap)
				{
					Found = TRUE;	save_x = x;	save_y = y;
				}
			}
			x += x_step;
		}
	}

	if (Found)
	{
		*new_x = save_x;	*new_y = save_y;
	}
	else {
#ifdef DEBUG
		fprintf(stderr,"Overlap occurs for %s at %ld,%ld (even with stepwise overlapcheck)\n",
			node->hostname, *new_x, *new_y);
#endif
	      }
}


	/* get layout changes for every node in the structure */

extern void layout_changes(factor)
float factor;				/* factor to multiply the forces with */
{
	struct hostentity * dummyhost1;	/* host to determine the changes for */
	struct hostentity * dummyhost2;	/* the host that interacts with dummyhost1 */
	struct hostentity * newhost;
	long int force_x, force_y;	/* force vectors for dummyhost1 */
	long int layout_x, layout_y;	/* total of force vectors for dummyhost1 */
	long int new_x;			/* new x_coordinate for the node */
	long int new_y;			/* new y_coordinate for the node */
	long int x_low, y_low;		/* use these to check on minimum coordinates */

	newhost = NULL;
	dummyhost1 = host;

	while (dummyhost1 != NULL)
	{
		layout_x = layout_y = 0;

#ifdef DEBUG
		fprintf(stderr,"layout_changes: working on %s\n",dummyhost1->hostname);
#endif

		dummyhost2 = host;

		while (dummyhost2 != NULL)
		{
			if (strcmp(dummyhost1->hostname,dummyhost2->hostname) != 0)
			{
				determine_force(dummyhost1, dummyhost2, &force_x, &force_y, factor);
				layout_x += force_x;
				layout_y += force_y;
			}

			dummyhost2 = dummyhost2->next;
		}

	/* divide total force_vectors by number of nodes to "prevend off-window-positioning" */

		layout_x /= number_of_nodes;
		layout_y /= number_of_nodes;

	/* determine the new coordinates for every node, using the old
	   coordinates and the grand totals of the force_vectors */

		new_x = dummyhost1->x_coord + layout_x;
		new_y = dummyhost1->y_coord + layout_y;

	/* check overlap with already changed nodes */

		check_overlap_layout(newhost, dummyhost1, &new_x, &new_y, layout_x, layout_y);

		/* make coordinates for this node known in a temp. structure */

		newhost = addnodestructure(newhost, dummyhost1->id, dummyhost1->hostname,
					   new_x, new_y,
					   dummyhost1->width, dummyhost1->height,
					   dummyhost1->link, dummyhost1->number, TRUE);

		dummyhost1 = dummyhost1->next;
	}

	/* after the complete cycle : make changes known to the original hoststructure
	   check minimum x- and y-coordinate, so the nodes stay reasonably close to (boundary,boundary) */

	x_low = 10 * WINDOWwidth;	y_low = 10 * WINDOWheight;

	host = newhost;
	while (host != NULL)
	{
		if (host->x_coord < x_low)
			x_low = host->x_coord;
		if (host->y_coord < y_low)
			y_low = host->y_coord;
		host = host->next;
	}

	if ( (x_low < 0) || (y_low < 0) || (x_low > WINDOWwidth) || (y_low > WINDOWheight) )
	{
		host = newhost;
		while (host != NULL)
		{
			host->x_coord = host->x_coord - x_low + boundary;
			host->y_coord = host->y_coord - y_low + boundary;
			host = host->next;
		}
	}
	host = newhost;
}
