/*******************************************************************************
 *
 *	shuffle.c
 *
 *	part of the layout program.
 *
 *
 *	Written by Gerard Leurs, NIKHEF-H, Amsterdam.
 *	January 1992.
 *	email: a89@nikhef.nl
 *
 *
 *	This part gets the coordinates of the nodes after the number of iterations that
 *	the forces have been determined. It solves overlapping (by "complex multiplying").
 *	It reduces the used space for all nodes (by "removing gaps" between nodes).
 *
 ******************************************************************************/


# include "layout.h"


	/* multiply coordinates in order to get no more overlap
	   if the overlap is too large : do NOT multiply (creates too large coordinates)
					 but simply move in several steps */

static void multiply_coordinates(width, xy_coord1, xy_coord2, is_x)
long int width;
int is_x;
long int xy_coord1, xy_coord2;
{
	struct hostentity * dummyhost1;
	float factor, tmp;
	float low_add, high_add;
	int Multiply = TRUE;	/* default. Only with large overlaps: false */
	int nr, slicepart;	/* parts of the width, for simply moving nodes */


	/* give an arbitrary factor if coordinates are equal ! */

	if (xy_coord1 != xy_coord2)
	{
		factor = (float) width / (xy_coord2 - xy_coord1);
		if (factor > 10.0)
			Multiply = FALSE;
	}
	else
		factor = 1.2;

	low_add = 0;
	if (Multiply)
		high_add = factor * ( xy_coord2 - xy_coord1 );
	else
		high_add = width;

#ifdef DEBUG
	if (Multiply)
		fprintf(stderr,"multiply_coordinates: multiplying\n");
	else
		fprintf(stderr,"multiply_coordinates: scaling\n");
#endif

	dummyhost1 = host_more;
	while (dummyhost1 != NULL)
	{
		if (is_x)
		{
			if (dummyhost1->x_coord < xy_coord1)
				dummyhost1->x_coord += low_add;
			else if (dummyhost1->x_coord > (xy_coord1 + width))
				dummyhost1->x_coord += high_add;
			else
			{
				if (Multiply)
				{
					tmp = factor * dummyhost1->x_coord;
					dummyhost1->x_coord += tmp;
				}
				else
				{
					slicepart = width / 10;
					for (nr=1; nr <= 10; nr++)
					{
						if ( (dummyhost1->x_coord <= (xy_coord1 + nr * slicepart))
						  && (dummyhost1->x_coord >= (xy_coord1 + (nr - 1) * slicepart)) )
      							dummyhost1->x_coord += (nr * slicepart);
					}
				}
			}
		}
		else
		{
			if (dummyhost1->y_coord < xy_coord1)
				dummyhost1->y_coord += low_add;
			else if (dummyhost1->y_coord > (xy_coord1 + width))
				dummyhost1->y_coord += high_add;
			else
			{
				if (Multiply)
				{
					tmp = factor * dummyhost1->y_coord;
					dummyhost1->y_coord += tmp;
				}
				else
				{
					slicepart = width / 5;
					for (nr=1; nr <= 5; nr++)
					{
						if ( (dummyhost1->y_coord <= (xy_coord1 + nr * slicepart))
						  && (dummyhost1->y_coord >= (xy_coord1 + (nr - 1) * slicepart)) )
      							dummyhost1->y_coord += (nr * slicepart);
					}
				}
			}
		}
		dummyhost1 = dummyhost1->next;
	}
}


	/* get minimum and maximum coordinates for all nodes */

extern void get_min_max_coord(nodestruct, low, high, is_x)
struct hostentity * nodestruct;
long int *low, *high;
int is_x;
{
	struct hostentity * dummyhost1;
	long int temp_height;


	/* get smallest coordinates used */

	if (is_x)
	{
		*low = 10 * WINDOWwidth;
		*high = - *low;
	}
	else
	{
		*low = 10 * WINDOWheight;
		*high = - *low;
	}

	dummyhost1 = nodestruct;
	while (dummyhost1 != NULL)
	{
		if (is_x)
		{
			if (dummyhost1->x_coord < *low)
				*low = dummyhost1->x_coord;
			if ((dummyhost1->x_coord + dummyhost1->width) > *high)
				*high = dummyhost1->x_coord + dummyhost1->width;

#ifdef DEBUG
			fprintf(stderr,"get_min_max_coord: [x] host %s at %ld %ld (width %ld, height %ld)\n",
				dummyhost1->hostname,
				dummyhost1->x_coord, dummyhost1->y_coord,
				dummyhost1->width, dummyhost1->height);
#endif
		}
		else
		{
			temp_height = get_height(dummyhost1);
			if (dummyhost1->y_coord < *low)
				*low = dummyhost1->y_coord;
			if ((dummyhost1->y_coord + temp_height) > *high)
				*high = dummyhost1->y_coord + temp_height;

#ifdef DEBUG
			fprintf(stderr,"get_min_max_coord: [y] host %s at %ld %ld (width %ld, height-all %ld)\n",
				dummyhost1->hostname,
				dummyhost1->x_coord, dummyhost1->y_coord,
				dummyhost1->width, temp_height);
#endif
		}

		dummyhost1 = dummyhost1->next;
	}

#ifdef DEBUG
	fprintf(stderr,"get_min_max_coord: min %ld max %ld\n", *low, *high);
#endif
}


	/* adjust coordinates slightly so the "lowest-node" lies close to (boundary,boundary) */

extern void adjust_coord(hoststr, is_x, is_final, bound)
struct hostentity *hoststr;
int is_x;			/* x values or yvalues */
int is_final;			/* called in proc. place_unlinked_nodes or not */
int bound;			/* boundary to use */
{
	struct hostentity * dummyhost1;
	long int min, max;		/* offset in x/y direction from (boundary,boundary) */


	if (is_x)
		get_min_max_coord(hoststr, &min, &max, TRUE);
	else
		get_min_max_coord(hoststr, &min, &max, FALSE);

	if (! is_final)
	{
		if (min < bound)
		{
#ifdef DEBUG
			fprintf(stderr,"adjust_coord: coordinates fixed\n");
#endif

			dummyhost1 = hoststr;
			while (dummyhost1 != NULL)
			{
				if (is_x)
					dummyhost1->x_coord += (bound - min);
				else
					dummyhost1->y_coord += (bound - min);
				dummyhost1 = dummyhost1->next;
			}
		}

		if ( ( (is_x) && (min > (WINDOWwidth  / 2)) )
		|| ( (! is_x) && (min > (WINDOWheight / 2)) ) )
		{
			dummyhost1 = hoststr;
			while (dummyhost1 != NULL)
			{
				if (is_x)
					dummyhost1->x_coord -= (min - (WINDOWwidth / 2 - bound));
				else
					dummyhost1->y_coord -= (min - (WINDOWheight / 2 - bound));
				dummyhost1 = dummyhost1->next;
			}
		}
	}
	else
	{
		dummyhost1 = hoststr;
		while (dummyhost1 != NULL)
		{
			if (is_x)
				dummyhost1->x_coord -= (min - bound);
			else
				dummyhost1->y_coord -= (min - bound);
#ifdef DEBUG
			fprintf(stderr,"adjust_coord: %s new_coordinates are %ld %ld\n",
				dummyhost1->hostname, dummyhost1->x_coord, dummyhost1->y_coord);
#endif
			dummyhost1 = dummyhost1->next;
		}
	}
}


	/* overlap between more_links_node and the more_links_node with two connections to
	   more_links_node */
	/* if there is an overlap between nodes we must know how large it is, so we
	   can multiply all coordinates of all nodes until there is no more overlap */
	/* this is a RECURSIVE routine */

extern int check_nodes_overlap()
{
	struct hostentity * dummyhost1;
	struct hostentity * dummyhost2;
	long int x1, y1, x2, y2;
	long int w1, h1, w2, h2;
	long int x_coord1, y_coord1, x_coord2, y_coord2;	/* for multiplying coordinates -> no more overlap */
	long int width_1, height_1;		/* dimensions of the "lowest node" */
	int count, nodecount;			/* needed to skip the already determined nodes */
	long int x_overlap, y_overlap;
	int Overlap = FALSE;
	long int xdir, ydir;
	int nr;


	fprintf(stderr,"Checking for overlap of nodes\n");

	/* adjust coordinates slightly so the "lowest-node" lies close to (boundary,boundary) */

	adjust_coord(host_more, TRUE, FALSE, boundary);	adjust_coord(host_more, FALSE, FALSE, boundary);

	/* determine existing overlap (and how large) */

	x_overlap = 0;		y_overlap = 0;		nodecount = 0;

	dummyhost1 = host_more;
	while (dummyhost1 != NULL)
	{
		nodecount ++;
		x1 = dummyhost1->x_coord;	y1 = dummyhost1->y_coord;	w1 = dummyhost1->width;
		h1 = get_height(dummyhost1);

		dummyhost2 = host_more;

		for (count=1; count <= nodecount ;count++)
			dummyhost2 = dummyhost2->next;

		while (dummyhost2 != NULL)
		{
		if (dummyhost1->number != dummyhost2->number)
		{
			x2 = dummyhost2->x_coord;	y2 = dummyhost2->y_coord;	w2 = dummyhost2->width;
			h2 = get_height(dummyhost2);

			/* overlap in x-coordinates (width) -> check y-coordinates */

			if ( ((x1 <= x2) && (x1 + w1 >= x2)) || ((x2 <= x1) && (x2 + w2 >= x1)) )
			{
				if ( (y1 <= y2) && ((y1 + h1) >= y2) )
				{                                                        
#ifdef DEBUG
	fprintf(stderr,"check_nodes_overlap: overlap for %s (%ld %ld %ld %ld) and %s (%ld %ld %ld %ld)\n",
		dummyhost1->hostname, x1, y1, w1, h1, dummyhost2->hostname, x2, y2, w2, h2);
#endif
					Overlap = TRUE;
					if ((y1 + h1 + 10 - y2) > y_overlap)
					{
						y_overlap = y1 + h1 - y2;
						y_coord1 = y1;	y_coord2 = y2;	height_1= h1 + 10;
					}
					if (x1 <= x2)
					{
						if ((x1 + w1 + 10 - x2) > x_overlap)
						{
							x_overlap = x1 + w1 - x2;
							x_coord1 = x1;	x_coord2 = x2;	width_1= w1 + 10;
						}
					}
					else
					{
						if ((x2 + w2 + 10 - x1) > x_overlap)
						{
							x_overlap = x2 + w2 - x1;
							x_coord1 = x2;	x_coord2 = x1;	width_1= w2 + 10;
						}
					}
				}
				else if ( (y2 <= y1) && ((y2 + h2) >= y1) )
				{
#ifdef DEBUG
	fprintf(stderr,"check_nodes_overlap: overlap for %s (%ld %ld %d %d) and %s (%ld %ld %ld %ld)\n",
		dummyhost1->hostname, x1, y1, w1, h1, dummyhost2->hostname, x2, y2, w2, h2);
#endif
					Overlap = TRUE;
					if ((y2 + h2 + 10 - y1) > y_overlap)
					{
						y_overlap = y2 + h2 - y1;
						y_coord1 = y2;	y_coord2 = y1;	height_1= h2 + 10;
					}
					if (x1 <= x2)
					{
						if ((x1 + w1 + 10 - x2) > x_overlap)
						{
							x_overlap = x1 + w1 - x2;
							x_coord1 = x1;	x_coord2 = x2;	width_1= w1 + 10;
						}
					}
					else
					{
						if ((x2 + w2 + 10 - x1) > x_overlap)
						{
							x_overlap = x2 + w2 - x1;
							x_coord1 = x2;	x_coord2 = x1;	width_1= w2 + 10;
						}
					}
				}
			}

		} /* end of if (nr1 != nr2) */

			dummyhost2 = dummyhost2->next;
                } /* end of while (dum2 != NULL) */
		dummyhost1 = dummyhost1->next;
	}

	/* fix the overlap for nodes */

	if (Overlap)
	{
		fprintf(stderr,"largest x_overlap for nodes at %ld (width %ld) and %ld\n",
			x_coord1, (width_1 - 10), x_coord2);
		fprintf(stderr,"largest y_overlap for nodes at %ld (height %ld) and %ld\n",
			y_coord1, (height_1 - 10), y_coord2);

		/* solve a tricky situation (can't do it by multiply_coordinates) */

		if ((x_coord1 == x_coord2) && (y_coord1 == y_coord2))
		{
			fprintf(stderr,"check_nodes_overlap: Both coordinates are exactly the same (%ld %ld).\n",
				x_coord1, y_coord1);

			dummyhost1 = host_more;
			while (dummyhost1 != NULL)
			{
				if ( (dummyhost1->x_coord == x_coord1)
				  && (dummyhost1->y_coord == y_coord1) )
				{
					xdir = ydir = 0;	nr = 0;

					dummyhost2 = host_more;
					while (dummyhost2 != NULL)
					{
						if ( (dummyhost1->number != dummyhost2->number)
						 && (linked_or_not(dummyhost1->number,dummyhost2->number)) )
						{
							xdir += dummyhost2->x_coord;
							ydir += dummyhost2->y_coord;
							nr ++;
						}
						dummyhost2 = dummyhost2->next;
					}

					xdir /= nr;	ydir /= nr;
					dummyhost1->x_coord += (dummyhost1->x_coord - xdir) / 25;
					dummyhost1->y_coord += (dummyhost1->y_coord - ydir) / 25;
				}
				dummyhost1 = dummyhost1->next;
			}
		}
		else
		{
			/* multiply all coordinates, so the largest overlap is fixed !! */

			multiply_coordinates(width_1,  x_coord1, x_coord2, TRUE);
			multiply_coordinates(height_1, y_coord1, y_coord2, FALSE);

			check_nodes_overlap();
		}
	}
	else
		fprintf(stderr,"There is NO overlap (anymore)\n");

	/* adjust coordinates slightly so the "lowest-node" lies close to (boundary,boundary) */

	adjust_coord(host_more, TRUE, FALSE, boundary);	adjust_coord(host_more, FALSE, FALSE, boundary);
}
