/****************************************************************/
/****************************************************************/
/*  $Header$   */
/****************************************************************/
/*  The model used for the network consists of a set of nodes   */
/*  or nodes connected to other nodes via interfaces, an    */
/*  interface can have up to two connections.  Both nodes and */
/*  interfaces are entities.                                    */
/*  MonetEntity classes can be networked in exactly the same way,    */
/*  this is useful because it defines which entities can be     */
/*  connected via an interface.  For instance we could prevent  */
/*  two ip hosts being connected together directly, by only     */
/*  defining an interface between an IP network and an IP host. */

/*  The make_interface and can_interface are a total mess. !!!  */
/****************************************************************/

#include "entityP.h"

/****************************************************************/
/*  Definitions  */
/****************************************************************/

/***************************************************************/
/*  Query for interfaces state  */
/***************************************************************/

extern MonetEntity from_interface (interface)
     MonetEntity interface;
{
  if (MonetEntity_subclass (MonetInterfaceClass, interface))
    return FROM (interface);

  MonetSysError ("from_interface:Not an interface [%s]",
	     MonetEntity_title (interface));
  return NULL;
}

extern MonetEntity to_interface (interface)
     MonetEntity interface;
{
  if (MonetEntity_subclass (MonetInterfaceClass, interface))
    return TO (interface);

  MonetSysError ("to_interface:Not an interface [%s]",
	     MonetEntity_title (interface));
  return NULL;
}

/***************************************************************/
/*  Can two objects be interfaced together ?  */
/***************************************************************/

extern MonetBool can_interface (interface, node)
     MonetEntity interface, node;
{
  MonetBool fromClass, toClass, fromFree, toFree;
  MonetEntity node_class;
  
  if (MonetEntity_isClass (node))
    node_class = node;
  else
    node_class = MonetEntity_classOf (node);
  if (MonetEntity_isInterfaceClass (interface)) {
    fromClass = MonetEntity_subclass (get_fromClass(interface), node_class);
    toClass   = MonetEntity_subclass (get_toClass(interface),   node_class);
    if (fromClass && toClass)
      return True;
    MonetPrintfError ("Interfaces of class [%s] cannot connect these two nodes",
		MonetEntity_title (interface));
    return False;
  } else if (MonetEntity_isInterface (interface)) {
    fromFree = (FROM(interface) == NULL);
    toFree   = (TO(interface) == NULL);
  } else {
    MonetSysError ("[%s] is not an interface", MonetEntity_title (interface));
    return False;
  }
  fromClass = MonetEntity_subclass (get_fromClass(interface), node_class);
  toClass   = MonetEntity_subclass (get_toClass(interface),   node_class);
  
  if (fromFree && fromClass) {
    if (!MonetEntity_isClass (node) && (TO(interface) == node)) {
      MonetPrintfError ("Cannot interface [%s] to itself",
			MonetEntity_title (node));
      return False;
    }
  } else if (toFree && toClass) {
    if (!MonetEntity_isClass (node) && (FROM(interface) == node)) {
      MonetPrintfError ("Cannot interface [%s] to itself",
			MonetEntity_title (node));
      return False;
    }
  } else if (!toFree && !fromFree) {
    MonetPrintfError ("[%s] is already connected to two nodes",
		      MonetEntity_title (interface));
    return False;
  } else {
    MonetPrintfError ("Cannot connect interface [%s] because of class conflict",
		MonetEntity_title (interface));
    return False;
  }
  return True;
}

static MonetBool interfaced (interface, node)
     MonetEntity interface, node;
{
  MonetEntity other;

  /* First check to see if already directly connected */
  if (FROM(interface) == node || TO(interface) == node)
    return True;
  if (FROM(interface) == NULL)
    other = TO(interface);
  else if (TO(interface) == NULL)
    other = FROM(interface);
  else
    return False;

  if (other == NULL)
    return False;


  /* !!! Temporary, disallow multiple joins */
  return False;
  /* NOTREACHED */
  
  /* !!! We allow nodes to be interfaced more than once !!! */
  /* Check the 'other' side of the interface is not interfaced to node via another interface */

}

extern MonetBool are_interfaced (sys1, sys2)
	MonetEntity sys1, sys2;
{
	if (MonetEntity_isInterface (sys1))
		return interfaced (sys1, sys2);
	else if (MonetEntity_isInterface (sys2))
		return interfaced (sys2, sys1);

	return False;
}

/***************************************************************/
/*  Connect and disconnect Systems and Interfaces */
/*  One version only for use in this library, it does not  */
/*  call trigger routines.  Public version does  */
/***************************************************************/

static MonetBool make_interface (interface, node)
     MonetEntity interface, node;
{
  MonetBool fromClass, toClass, fromFree, toFree;
  String title_interface, title_node; /*!!!*/
  MonetBool already_in = False;

  if (interface == NULL) {
    MonetSysError ("make_interface: interface is NULL");
    return False;
  }
  if (node == NULL) {
    MonetSysError ("make_interface: node is NULL");
    return False;
  }
  title_interface = MonetEntity_title(interface);
  if (MonetEntity_isInterface (interface) == False &&
      MonetEntity_isInterfaceClass (interface) == False) {
    MonetSysError ("[%s] is not an interface", title_interface);
    return False;
  }
  title_node      = MonetEntity_title(node);
  
  if (MonetEntity_isClass (interface))
    if (MonetEntity_isClass (node)) {
      fromClass = True;
      toClass   = True;
    } else {
      MonetPrintfError ("Cannot interface a class [%s] to an instance [%s]",
		  title_interface, title_node);
      return False;
    }
  else if (MonetEntity_isClass (node)) {
    MonetPrintfError ("Cannot interface a class [%s] to an instance [%s]",
		title_node, title_interface);
    return False;
  } else {
    fromClass = MonetEntity_subclass (get_fromClass(interface),
				      MonetEntity_classOf(node));
    toClass   = MonetEntity_subclass (get_toClass(interface),
				      MonetEntity_classOf(node));
  }
  fromFree  = FROM(interface) == NULL;
  toFree    = TO(interface) == NULL;
  
  if (fromFree && fromClass) {
    if (TO(interface) == node) {
      if (MonetEntity_isClass (node)) {
	already_in = True;
      } else {
	MonetSysError ("Cannot interface [%s] to itself", title_node);
	return False;
       }
    }
    LVAL_FROM(interface) = (caddr_t) node;
  } else 	if (toFree && toClass) {
    if (FROM(interface) == node) {
      if (MonetEntity_isClass (node)) {
	already_in = True;
      } else {
	MonetSysError ("Cannot interface [%s] to itself", title_node);
	return False;
      }
    }
    LVAL_TO(interface) = (caddr_t) node;
  } else if (!toFree && !fromFree) {
    MonetPrintfError ("[%s] already has two interfaces", title_interface);
    return False;
  } else {
    MonetPrintfError ("Can't interface [%s] to objects of class [%s]",
		      title_interface,
		      MonetEntity_title (MonetEntity_classOf(node)));
    return False;
  }
  if (! already_in)
    MonetEntityList_insert (MonetEntity_interfaces (node), interface);
  return True;
}

extern MonetBool MonetEntity_make_interface (interface, node)
     MonetEntity interface, node;
{
  if (make_interface (interface, node)) {
    if (FROM(interface) == node) {
      MonetMask_SET(MonetAttribute_FROM_NUM, &MASK(interface));
    } else if (TO(interface) == node) {
      MonetMask_SET(MonetAttribute_TO_NUM, &MASK(interface));
    }
    MonetEntity_change_trigger (interface);
    return True;
  }
  return False;
}

/***************************************************************/
/*  If node and interface are linked then disconnect them.     */
/***************************************************************/

static MonetBool unmake_interface (interface, node)
     MonetEntity interface, node;
{
  MonetEntityList list;
  if (node == NULL) {
    return False;
  }
  if (list = MonetEntity_interfaces (node)) {
    if (FROM(interface) == node) {
      LVAL_FROM(interface) = NULL;
      MonetMask_SET(MonetAttribute_FROM_NUM, &MASK(interface));
    } else if (TO(interface) == node) {
      LVAL_TO(interface) = NULL;
      MonetMask_SET(MonetAttribute_TO_NUM, &MASK(interface));
    } else {
      MonetSysError ("Not interfaced");
      return False;
    }
    MonetEntityList_extract (list, interface);
    MonetEntity_change_trigger (interface);
  }
  return True;
}

extern MonetBool MonetEntity_unmake_interface (interface, node)
     MonetEntity interface, node;
{
  unsigned num;
  if (FROM(interface) == node) {
    num = MonetAttribute_FROM_NUM;
  } else if (TO(interface) == node) {
    num = MonetAttribute_TO_NUM;
  }
  if (unmake_interface (interface, node)) {
    MonetMask_SET(num, &MASK(interface));
    MonetEntity_change_trigger (interface);
    return True;
  }
  return False;
}

/***************************************************************/
/***  Disconnects all the interfaces of an object ****/
/*  !!! Should be called something else */
/***************************************************************/

extern void free_interfaces (node)
	MonetEntity node;
{
  MonetEntity interface;
  MonetEntityList list;

  if (MonetEntity_isClass (node)) {
    return;
  } else if (MonetEntity_isInterface (node)) {
    unmake_interface (node, FROM(node));
    unmake_interface (node, TO(node));
    return;
  }
  if (list = MonetEntity_interfaces (node)) {
    FOR_ENTITYLIST (interface, list) {
      if (FROM(interface) == node) {
	MonetMask_SET(MonetAttribute_FROM_NUM, &MASK(interface));
	LVAL_FROM(interface) = NULL;
      } else if (TO(interface) == node) {
	MonetMask_SET(MonetAttribute_TO_NUM, &MASK(interface));
	LVAL_TO(interface) = NULL;
      }
      MonetEntity_change_trigger (interface);
    } END_FORLIST;
    MonetEntityList_reset (list);
  }
}


extern MonetBool MonetEntity_make_links (interface, from, to)
     MonetEntity interface, from, to;
{
  MonetBool change = False;

  if (FROM(interface) != from) {
    unmake_interface (interface, FROM(interface));
    make_interface (interface, from);
    MonetMask_SET(MonetAttribute_FROM_NUM, &MASK(interface));
    change = True;
  }
  if (TO(interface) != to) {
    unmake_interface (interface, TO(interface));
    make_interface (interface, to);
    MonetMask_SET(MonetAttribute_TO_NUM, &MASK(interface));
    change = True;
  }
  if (change) {
    MonetEntity_change_trigger (interface);
    return True;
  }
  return False;
}

/*****************************************************************/
/* Set interfaces with a string, no difference between these     */
/* These are called when parsing the source files */
/*****************************************************************/

static String access_interface (interface, id, op, format,
				null_default, node_name)
     MonetEntity interface;
     unsigned id;
     MonetBool null_default;
     AttributeOperation op;
     AttributeFormat format;
     String node_name;
{
  MonetEntity node;

  node = (MonetEntity) MONETGETPOINTER(interface, id);
  if (op == attribute_init) {
    MONETGETPOINTER(interface, id) = (caddr_t) NULL;
  } else if (op == attribute_free || op == attribute_unset) {
    unmake_interface (interface, node);
    MonetMask_SET(id, &MASK(interface));
  } else if (op == attribute_set) {
    MonetEntity new_node;
    if (! node_name || *node_name == NULL) {
      unmake_interface (interface, node);
      MonetMask_SET(id, &MASK(interface));
    } else {
      new_node = MonetEntity_find_or_create (node_name);
      if (node == new_node) {
	return NULL;
      } else {
	unmake_interface (interface, node);
	make_interface (interface, new_node);
	MonetMask_SET(id, &MASK(interface));
      }
    }
  } else if (node) {
    return MonetEntity_title (node);
  }
  if (null_default && node == NULL)
    return NULL;
  return "";
}

/*****************************************************************/
/*  Systems have lists of all the interfaces they are connected  */
/*  to, interfaces do not  */
/*****************************************************************/

static String access_interface_list (entity, id, op, format,
				     null_default, node_name)
     MonetEntity entity;
     unsigned id;
     MonetBool null_default;
     AttributeOperation op;
     AttributeFormat format;
     String node_name;
{
  MonetEntityList list;

  if (MonetEntity_isInterface (entity)) {
    entity->interfaces = NULL;
    return NULL;
  }
  if (op == attribute_init) {
    entity->interfaces	= MonetEntityList_create ();
  } else if (op == attribute_free || op == attribute_unset) {
    free_interfaces (entity);
    MonetEntityList_free (MonetEntity_interfaces(entity));
    entity->interfaces	= NULL;
  } else if (op == attribute_set) {
  }
  if (list = MonetEntity_interfaces (entity)) {
    MonetEntity interface;
    static char buffer [2048]; /* !!! */
    char * cp = buffer;
    FOR_ENTITYLIST (interface, list) {
      cp = MonetStringCopyToEnd (cp, MonetEntity_title (interface));
      *cp = '\n'; ++cp;
   }  END_FORLIST;
    if (cp > buffer) {
      *cp = '\0';
      return buffer;
    }
  }

  return NULL;
}

/****************************************************************/
/* init and finish  */
/****************************************************************/

extern void monet_db_init_interface ()
{
  MonetType interface, interface_list;
  interface = MonetType_create
    ("InterfaceLink", False, access_interface, NULL,
     "Link from an interface to a system");
  interface_list = MonetType_create
    ("InterfaceList", True, access_interface_list, NULL,
     "Links from a system to an interface");
  (void) MonetAttribute_create ("interfaces",   MonetAttribute_INTERFACE_NUM,
				interface_list, "");
  (void) MonetAttribute_create ("from",   MonetAttribute_FROM_NUM,
				interface, "");
  (void) MonetAttribute_create ("to",     MonetAttribute_TO_NUM,
				interface, "");
}

extern void monet_db_finish_interface ()
{
}

/****************************************************************/
/****************************************************************/
/*  $Header$   */
/****************************************************************/
/****************************************************************/


