/**********************************************************/
/**********************************************************/
/*  $Header$  */
/**********************************************************/
/*  This is where all the high level comms is done.       */
/*  This module sits on top of channel.c which does       */
/*  buffered asyncronous communcations between two        */
/*  processes.                                            */
/*                                                        */
/*  Ask another program to monitor objects, rather than   */
/*  doing it locally.                                     */
/*  Each monitor is specified by its address and by a     */
/*  pattern.  Rules added later have priority over those  */
/*  added earlier.                                        */
/*                                                        */
/*  Is also responsible for handling probe events.  When  */
/*  redirected the probing is switched off.  When not     */
/*  redirected the probing begins.                        */
/**********************************************************/

#include "entityP.h"
#include "requestP.h"
#include <monet/msh/Channel.h>
#include <monet/msh/Script.h>
#include <netdb.h>

#include <sys/param.h>  /*  For MAXHOSTNAMELEN  */

/***************************************************************************/
/*  Lookup table, this should be some kind of lexical automaton, with !!!  */
/*  end nodes pointing to (host,port) or NOTFOUND.  As this is quite  !!!  */
/*  hard to do, a simple table will be used as a prototype.           !!!  */
/***************************************************************************/

struct monet_redirect {
  long delay;  /* Delay between trying to reconnect  */
  String pattern;
  String host;
  int port;
  Channel channel;
  unsigned ping_id;
  MonetRedirect next;
  unsigned long int last_checked;
  MonetEntity loading_entity;
  MonetEntityList entitys;
};
#define MAXPINGID (unsigned) (0xffffffff)

static MonetRedirect   RedirectList;
static MonetBool       Disable;
static MonetEntityList BaseRule;
static MonetBool       IsServer;
static long	       Monitor_port;
static MonetBool       ProbingEnabled;
static String          Monitor_address	;

static MonetStr ProbeStr, StartProbeStr;

/***************************************************************************/
/* Put on to base rule  */
/***************************************************************************/

static void base_rule (entity)
     MonetEntity entity;
{
  PINGID(entity) = 0;
  entity->redirect = NULL;
  MonetEntityList_insert (BaseRule, entity);
}

/***************************************************************************/
/*  Send message  */
/***************************************************************************/

static void send_message (rule, message)
     MonetRedirect rule;
     char * message;
{
  if (rule && rule->channel) {
    (void) Channel_send_record (rule->channel, message);
  }
}

/***************************************************************************/
/* Match address against list of regular expressions */
/***************************************************************************/

static MonetRedirect match_rule (title)
     char * title;
{
  MonetRedirect rule;
  FORALL(rule, RedirectList, next) {
    MONETDEBUG(10, ("Matching pattern [%s] against [%s]", rule->pattern,title));
    if (MonetMatchWildcard (title, rule->pattern))
      return rule;
  }
  return NULL;
}

/******************************************************************/
/*  Start a probe event on an entity  */
/******************************************************************/

static void start_probe_event (entity)
     MonetEntity entity;
{
  if (ProbingEnabled) {
    /* !!! should not do a string conversion here  */
    MONETDEBUG(1, ( "start-probe from [%s]", MonetEntity_title (entity)));
    (void) MonetEntity_do_action (entity, "start-probe", 0);
  }
}

static void remove_probe (entity)
     MonetEntity entity;
{
  MONETDEBUG(1, ( "Removing probe from [%s]", MonetEntity_title (entity)));
  MonetEvent_free_entity_by_name (entity, StartProbeStr);
  MonetEvent_free_entity_by_name (entity, ProbeStr);
}

/******************************************************************/
/*  Check for cycles between the agents  */
/******************************************************************/

extern MonetBool MonetRedirect_check_cycle (host, port, pid)
     String host;
     int port, pid;
{
  if (MonetStringsEqual(host, Monet_host_name ()) &&
      port == Monitor_port && pid == getpid ()) {
    return True;
  }
  return False;
}

/******************************************************************/
/*  Stop redirecting requests for a entity  */
/******************************************************************/

static void delete_entity_redirect (entity)
	MonetEntity entity;
{
  if (entity->redirect->channel) {
    char buffer [24];
    (void) sprintf (buffer, "delete%u", PINGID(entity));
    (void) Channel_send_record (entity->redirect->channel, buffer);
  }
  (void) MonetEntityList_extract (entity->redirect->entitys, entity);
  PINGID(entity) = 0;
  entity->redirect = NULL;
  start_probe_event (entity);
}


/******************************************************************/
/*  Stop redirecting requests for a entity  */
/******************************************************************/

static void stop_redirect (entity)
	MonetEntity entity;
{
  if (entity->redirect) {
    delete_entity_redirect (entity);
    base_rule (entity);
  }
}

/*****************************************************************
  Send a request for monitoring to the agent.
  If channel not connected then try doing monitoring locally.
  *****************************************************************/

static MonetBool send_request (entity)
     MonetEntity entity;
{
  MonetRedirect rule = entity->redirect;

  if (rule->channel) {
    char buffer[24];
    (void) sprintf (buffer, "request%u ", PINGID(entity));
    MONETDEBUG(10, ("Sending Redirect Request [%s]#%d to monitor %d",
		MonetEntity_title (entity), PINGID(entity), rule->channel));
    return (Channel_send (rule->channel, buffer) &&
	    Channel_send_record (rule->channel, MonetEntity_title (entity)));
  } else {
    start_probe_event (entity);
  }
}

/********************************************************************/
/*  Try to find a matching redirection rule                         */
/*  If a match is found then all probing events are removed.  When  */
/*  was redirected and no match found it starts a probe event.      */
/********************************************************************/

static MonetRedirect try_redirect (entity)
     MonetEntity entity;
{
  MonetRedirect rule;
  String area;
  MonetStr area_str;

  if (area_str = (MonetStr) MONETAREA(entity)) {
    area = MonetStr2String (area_str);
  } else {
    area = MonetEntity_title (entity);
  }

  /* Match address against list of regular expressions */
  rule = match_rule (area);

  if (entity->redirect == rule) {
    /* -- Already associated with this redirect */
    start_probe_event (entity);
    return rule;
  }
  if (rule == NULL) {
    /* -- Previous rule no longer applies, should this happen ? */
    stop_redirect (entity);
    start_probe_event (entity);
    return NULL;
  }
  if (rule->ping_id == MAXPINGID) {
    /* !!! Should recover from this condition, ping id's ought to be long */
    MonetPrintfError
      ("Sorry no more Ping ID's available, can't ping any more entities");
    /*  Start probe event  */
    return NULL;
  }
  if (entity->redirect) {
    /* -- Previous rule no longer applies, since it has previous rule
       -- is should also have no probe event waiting for it */
    stop_redirect (entity);
  } else {
    /* -- Stop probe event  */
    /* !!! Should not know about redirects */
    remove_probe (entity);
  }
  (void) MonetEntityList_extract (BaseRule, entity);
  PINGID(entity) = rule->ping_id;
  ++ rule->ping_id;
  MonetEntityList_insert (rule->entitys, entity);
  entity->redirect = rule;

  (void) send_request (entity);

  return rule;
}

/******************************************************************/
/*  Stop redirecting requests for a entity  */
/******************************************************************/

extern void RedirectMonetEntity_delete (entity)
	MonetEntity entity;
{
  if (entity->drawing || entity->requests) {
    /* Only delete if not referenced by something else */
    /* this is actually a hack since this library should !!!
       know nothing about Drawings.  !!!  */
  } else {
    if (entity->redirect) {
      delete_entity_redirect (entity);
    }
    /*  Kill the probe procedure  */
    remove_probe (entity);
    (void) MonetEntityList_extract (BaseRule, entity);
  }
}

/****************************************************************************/
/*  Register an entity for redirection  */
/*  Must use RedirectMonetEntity_delete before destroying the entity  */
/****************************************************************************/

extern void RedirectMonetEntity_add (entity)
     MonetEntity entity;
{
  if (entity->drawing || entity->requests) {
    /*  Already added, !!! not the way to do things */
  } else {
    base_rule (entity);
    (void) try_redirect (entity);
  }
}

/**********************************************************/
/*  Parse/clear all redirects      */
/**********************************************************/

MONETCALLBACK(do_clear_redirects)
    clear_redirects ();
MONETEND

/**********************************************************/
/*  Allow probing      */
/**********************************************************/

MONETCALLBACK(do_probing_enabled)
     MonetEntity probed;
     if (argc == 1) {
       /*  -- Just querying not changing  */
     } else if (argv[1] && !ProbingEnabled) {
       ProbingEnabled = True;
       FOR_ENTITYLIST (probed, BaseRule) {
	 start_probe_event (probed);
       } END_FORLIST;
     } else if (!argv[1] && ProbingEnabled) {
       ProbingEnabled = False;
       FOR_ENTITYLIST (probed, BaseRule) {
	 remove_probe (probed);
       } END_FORLIST;
     } else {
       /* -- State not changed */
     }
     if (!ProbingEnabled) {
       result = NULSTR;
     }
MONETEND

/*****************************************************************************/
/*  Create a redirect  */
/*****************************************************************************/

MONETCALLBACK(do_redirect)
     String host, regexp;
     int port;
     MonetRedirect rule;

     if (!(host = argv[1]) || !(regexp = argv[3])) {
       MonetPrintfError ("Hostname or pattern is null");
       result = NULL;
       RETURNSC;
     }
     port = MonetString2int (argv[2]);
     

     if (rule = create_redirect (host, port, regexp)) {
       if (argc == 4) {
	 rule -> delay = MonetString2long (argv[4]);
       }
/*    MONETDEBUG(10, ("Will redirect [%s] to [%s+%u]", regexp, host, port)); */
       MonetPrintfInform ("Redirect [%s] to [%s+%u]",
			  regexp, rule->host, rule->port);
     } else {
       result = NULL;
     }
MONETEND

/****************************************************************************/
/*  List redirects */
/****************************************************************************/

MONETCALLBACK(do_help_redirects)
     MonetRedirect rule;
     int n = 0;

     FORALL (rule, RedirectList, next) {
       ++ n;
       MonetPrintf ("%d.", n);
       MonetPrintf ("\thost    [%s+%d] delay [%lu]",
		    rule->host, rule->port, rule->delay);
       if (rule->channel)
	 MonetPrintfnl (" Connected,  writes buffered [%u]",
			Channel_write_buffered (rule->channel));
       else
	 MonetPrintfnl (" Not connected");
       MonetPrintfnl ("\tpattern [%s]", rule->pattern);
       MonetPrintfnl ("\tentities [%d]",
		      MonetEntityList_sizeOf(rule->entitys));
       if (argv[1]) {
	 MonetEntity e;
	 FOR_ENTITYLIST (e, rule->entitys) {
	   MonetPrintfnl ("%u [%s]", PINGID(e), MonetEntity_title (e));
	 } END_FORLIST;
       }
     }
     MonetPrintPrettyCount ("Not redirected", "entity",
			    MonetEntityList_sizeOf (BaseRule));
     MonetPrintPrettyCount ("Found", "redirect", n);
MONETEND

/****************************************************************************/
/*  Enable/Disable redirects */
/****************************************************************************/

MONETCALLBACK(do_disable_redirects)
     disable_redirects ();
MONETEND

MONETCALLBACK(do_enable_redirects)
     enable_redirects ();
MONETEND

MONETCALLBACK(do_redirect_reconnect)
  if (Disable == False) {
    reconnect_redirects ();
  }
MONETEND

/****************************************************************************/
/*  Handle messages sent from the pinger process  */
/*  !!! Not very efficient lookup, should have 'keyed' objectlists !!! */
/*  !!! Should update each system in one go, not for each attribute */
/*  !!! This should really be put in the 'status' library.                  */
/* Read message, convert it into a status message */
/* Set the approriate status bits in the entity.     */
/* Arrange for 'request' to return this status.   */
/* Possibly relay this new information straight   */
/* back the channels.                              */
/****************************************************************************/

static MonetBool handle_response (chan, result)
     Channel chan;
     String result;
{
  MonetRedirect rule = (MonetRedirect) Channel_data (chan);

  MONETDEBUG(30, ( "Received [%s] from redirect server", result));
  if (rule == NULL) {
    MonetSysError ("handle_reponse: rule is NULL");
    return False; /* -- Disconnect */
  } else if ( ! MonetEntity_load_line (&(rule->loading_entity), result)) {
    /*  Some kind of error */
  }
  return True;
}

/***************************************************************************/
/*  When a channel disconnects we must set all the Redirects connected     */
/*  to it to NULL.  The redirects themselves are not destroyed, just       */
/*  disabled until a new channel can be created.                           */
/***************************************************************************/

static void disconnect (chan)
     Channel chan;
{
  caddr_t data = Channel_data (chan);
  MonetRedirect r, rule;

  rule = (MonetRedirect) data;

  /* If any of the other redirects share this channel then
     disconnect them too.  */
  FORALL(r, RedirectList, next)
    if (r->channel == rule->channel)
      r->channel = NULL;
  rule->channel = NULL;
  MonetPrintfInform ("Redirect [%s+%d] disconnected", rule->host, rule->port);
}

/***************************************************************************/
/*  Try to connect to the server specified for 'rule'  */
/***************************************************************************/

static void open_redirect_channel (rule)
     MonetRedirect rule;
{
  MonetRedirect l;

  if (rule->channel)
    return;
  FORALL(l, RedirectList, next)
    if (l != rule && EQS(rule->host, l->host) && rule->port==l->port) {
      rule->channel = l->channel;
      return;
    }
  rule->channel =  Channel_create_client (rule->port, rule->host, False,
					 handle_response, disconnect,
					 (caddr_t) rule);
  if (rule->channel) {
    MonetEntity entity;
    MONETDEBUG(10, ("Channel created"));
    Redirect_say_hello (rule);
    FOR_ENTITYLIST (entity, rule->entitys) {
      (void) send_request (entity);
    } END_FORLIST;
  } else {
    MONETDEBUG(10, ("Channel not created"));
  }
}

/***************************************************************************/
/*  Initiate transfer  */
/***************************************************************************/

extern void Redirect_say_hello (rule)
     MonetRedirect rule;
{
  if (rule && rule->channel) {
    char buffer[MAXHOSTNAMELEN + 64];
    (void) sprintf (buffer, "hello %s %u %u",
		    Monet_host_name (), Monitor_port, getpid ());

    send_message (rule, buffer);
  }
}

/***************************************************************************/
/*  Try to attach to any disconnected servers  */
/***************************************************************************/

extern void reconnect_redirects ()
{
  MonetRedirect rule;

  if (Disable)
    return;
  FORALL (rule, RedirectList, next)
    open_redirect_channel (rule);
}		

/***************************************************************************/
/*  Temporarily close connections  */
/***************************************************************************/

extern void disable_redirects ()
{
  MonetRedirect rule, l;

  FORALL (rule, RedirectList, next) {
    Channel channel = rule->channel;
    if (channel) {
      Channel_shutdown (channel);
      FORALL (l, RedirectList, next)
	if (channel==l->channel)
	  l->channel=NULL;
    }
  }
  Disable = True;
}

extern void enable_redirects ()
{
  if (Disable) {
    Disable = False;
    reconnect_redirects ();
  }
}

/********************************************************************/
/*  Check to see if a entity is redirected elsewhere.               */
/*  A redirect applies to this object but there is no connection    */
/*  !!! Should maybe use this routine to check for new matches.     */
/********************************************************************/

extern MonetBool isRedirected (entity)
     MonetEntity entity;
{
  if (entity && entity->redirect && entity->redirect->channel)
    return True;

  return False;
}

/****************************************************************************/
/*  Reset/Clear all redirects  */
/****************************************************************************/

extern void clear_redirects ()
{
  while (RedirectList != NULL)
    free_redirect (RedirectList);
}

/***************************************************************/

/***************************************************************/

extern void Redirect_get_info (rule,host, port, pattern)
     MonetRedirect rule;
     String *host, *pattern;
     int *port;
{
  if ( ! rule) {
    MonetSysError ("rule is NULL");
     *host = NULL;
    *pattern = NULL;
    return;
  }
  *host    = rule->host;
  *pattern = rule->pattern;
  *port    = rule->port;
}

/*****************************************************************/
/*  Create basic redirect  */
/*****************************************************************/

static MonetRedirect create_basic (host, port, pattern)
     String host, pattern;
     int port;
{
  MonetRedirect redirect;
	
  redirect = MonetNew (struct monet_redirect);
  redirect -> delay   = 0;
  redirect -> pattern = MonetNewString (pattern);
  redirect -> host    = MonetNewString (host);
  redirect -> port    = port;
  redirect -> channel = NULL;
  redirect -> next    = RedirectList;
  redirect -> ping_id = 1;
  redirect -> entitys = MonetEntityList_create ();
  redirect -> loading_entity = NULL;
  RedirectList   = redirect;
  
  /*  Should look through all entities to see if any match
      this redirect rule  */

  return redirect;
}

/***************************************************************/
/*  Add and remove a redirect patern  */
/***************************************************************/

extern MonetRedirect create_redirect (host, port, pattern)
     String host, pattern;
     int port;
{
  MonetRedirect redirect;
  struct hostent *hp;
  MonetEntity entity;

  /* Try to connect to remote monitor and say hello */
  /* Add the regexp and new connection to a table   */

  if (! host || ! *host) {
    MonetPrintfError ("Failed to create redirect to [%s+%d]", host, port);
    return NULL;
  }
  if ((hp = gethostbyname (host)) == NULL) {
    MonetPrintfError ("redirect: No such host [%s]", host);
    return NULL;
  }
  host = hp->h_name;
  if (EQS(host, Monet_host_name ()) && MonetRequest_port () == port) {
    MonetPrintfWarning ("Cannot create a redirect to self [%s+%d][%s]",
			host, port, pattern);
    return NULL;
  } else if (redirect = create_basic (hp->h_name, port, pattern)) {
    /*  -- look through all entities in BaseRule to see
	-- if any match this redirect rule  */
    FOR_ENTITYLIST (entity, BaseRule) {
      try_redirect (entity);
    } END_FORLIST;
    /*  get a channel for this rule  */
    open_redirect_channel (redirect);
  } else {
    MonetPrintfInform ("Couldn't create redirect to monitor at [%s+%d]",
		       host, port);
  }
  return redirect;
}

/*****************************************************************/
/*  Is a redirct actually connected to a channel  */
/*****************************************************************/

extern MonetBool redirect_connected (rule)
     MonetRedirect rule;
{
  return (rule->channel != NULL);
}

/*****************************************************************/
/*  Free a redirect rule                                         */
/*****************************************************************/

extern void free_redirect (rule)
     MonetRedirect rule;
{
  MonetEntity entity;

  MonetPrintfInform ("Removing redirect [%s+%u]", rule->host, rule->port);

  /*  Remove 'rule' from 'RedirectList'
      do this now so that the FORALL(r, RedirectList, next)
      will work.  */
  LIST_DEL (MonetRedirect, rule, RedirectList, next);

  /*  Notify server by clearing requests !!! */
  /*  !!! This is where an acknowledgement is needed, else */
  /*  !!! we could get answers to queries just deleted */
  if (rule->channel) {
    MonetRedirect r;
    if (Channel_send_record (rule->channel, "clear")) {
      /* A convoluted way of only closing a connection to a server */
      /* when there are no redirects to it. */
      FORALL(r, RedirectList, next)
	if (r->channel == rule->channel) {
	  rule->channel = NULL;
	  break;
	}
      if (rule->channel) {
	Channel_shutdown (rule->channel);
      }
    }
  }

  /*  Try to find an alternative redirect for each entity  */
  FOR_ENTITYLIST (entity, rule->entitys) {
    base_rule (entity);
    try_redirect (entity);
  } END_FORLIST;
  MonetEntityList_reset (rule->entitys);
  MonetEntityList_free (rule->entitys);
  MonetFree (rule->pattern);
  MonetFree (rule->host);
  MonetFree (rule);
}

/*****************************************************************/
/* Init and Finish */
/*****************************************************************/

extern void monet_finish_redirect ()
{
  clear_redirects ();
  MonetEntityList_reset (BaseRule);
  MonetEntityList_free (BaseRule);
  MonetStr_derefer (ProbeStr);
  MonetStr_derefer (StartProbeStr);
}

extern void monet_init_redirect (is_server)
     MonetBool is_server;
{
  RedirectList = NULL;
  Disable = False;
  BaseRule = NULL;
  IsServer = False;
  Monitor_port = 0L;
  ProbingEnabled = False;
  Monitor_address = NULSTR;


  ProbeStr      = MonetStr_refer ("probe");
  StartProbeStr = MonetStr_refer ("start-probe");
  IsServer        = is_server;
  RedirectList  = NULL;
  BaseRule      = MonetEntityList_create ();

  (void) MonetMember_module_class ("redirect");

  MonetMember_create ( "help",		do_help_redirects,
		      "?bool",
		      "list all current redirects, \
if the second argument is given a full list of entities is printed.");
  MonetMember_create ("create",  do_redirect,
		      "host port pattern ?reconnect_delay",
		      "Set up a redirect, this is an instruction saying that \
any entities matching 'pattern' should not be probed locally but should be \
passed on to the agent specified by <host+port>");
  MonetMember_create ("clear",  do_clear_redirects,
		      "",
		      "Clear all redirects");
  MonetMember_create ("enable",  do_enable_redirects,
		      "",
		      "Enable traffic between client and server");
  MonetMember_create ("reconnect",  do_redirect_reconnect,
		      "",
		      "Attempt to reconnect to all disconnected agents");
  MonetMember_create ("disable",  do_disable_redirects,
		      "",
		      "Disable redirect traffic between this program and any \
agents it is redirecting to");
  MonetMember_create_write ( "host", &Monitor_address,
			    "Host name of default monitoring agent (monetd)");
  Monitor_address         = MonetNewString ("localhost");

  MonetMember_create_write_long ( "port", &Monitor_port,
				 "Tcp port of monetd server");
  Monitor_port = 2567;

  MonetMember_create
    ("probing-enabled", do_probing_enabled,
     "?flag",
     "Enables/disables probing, if enabled then tries to execute the \
'start-probe' member of all non-redirected entities.  If disabled it \
removes the 'probe' events for these entities.");
  ProbingEnabled = False;

  MonetConfig_create ("-probe", "redirect->probing-enabled", NULL,
		      "Enable probing by this program");
  MonetConfig_create ("-a", "redirect->host", NULL,
		      "Address of monitoring agent");
  MonetConfig_create ("-p", "redirect->port", NULL,
		      "Port number of monitoring agent");

}

/**********************************************************/
/**********************************************************/
/*  $Header$  */
/**********************************************************/
/**********************************************************/

