#ifndef MONET_DRAWLIB_H_
#define MONET_DRAWLIB_H_

#include <monet/msh/General.h>

/***********************************************************************

  At the top level is a library for drawing various basic shapes,
  each shape has one or maybe two labels and can have a set of
  extra attributes such as thick border, symbols and colours/shapes.

  The shapes library sits on top of a simple but generic drawing library
  which provides scaled graphics and text.

  This drawing library sits on top of a set of drivers, one for
  producing postscript, one for X11 and one giving a simple text format.

                   Shape
                     |
                   Draw
                  /  |  \
        Postscript  X11  Text

  Weaknesses are that the draw library is not that generic and the
  shapes library is very ungeneric, it should be possible for the user
  to design their own shapes.

  The shapes library works in two stages, first you pass the details
  of your drawing to a routine 'Shape_prepare' which may adjust them
  a bit.  After this you call 'Shape_draw' to display it.

***********************************************************************/


/****************************************************************
  Pixel height and widths for strings 
  6x8 is for the X font 5x8, which is the minimally readable   
  one.  Maybe this should not be hard wired and should be      
  different for postscript - but such is life.  
  A width of 6 for 5x8 is because of the one pixel seperator  
  ********************************************************************/

#define DrawCharHeight 8
#define DrawCharWidth  6

/***********************************************************************/
/* Shade can mean, dashing, grey scale, or colours */
/***********************************************************************/


typedef enum {
	shade_0,
	shade_1,
	shade_2,
	shade_3,
	shade_4,
	shade_5
#define NUM_SHADE 6
	}
	MonetShade;
#define shade_background  shade_0
#define shade_dark   shade_4
#define shade_grey   shade_2
#define shade_heavy  shade_5

typedef unsigned BitmapId;
/***********************************************************************/

/***********************************************************************/

typedef struct drawLibrary {
	MonetBool is_xor;
	void (* end_draw)
	  PR((void));
	void (* draw_percentage_line)
	  PR((MonetShade, Position, Position,
	      Position, Position, Position, Position));
	void (* draw_string)
	  PR((MonetShade, Position, Position, Dimension, Dimension,
	      String label, unsigned len));
	void (* draw_boxed_string)
	  PR((MonetShade, Position, Position, Dimension, Dimension,
	      String label, unsigned len));
	void (* draw_line)
	  PR((MonetShade, Position, Position, Position, Position));
	void (* draw_elipse)
	  PR((MonetShade, Position, Position, Dimension, Dimension));
	void (* draw_rectangle)
	  PR((MonetShade, Position, Position, Dimension, Dimension));
	void (* draw_roundedRectangle)
	  PR((MonetShade, Position, Position, Dimension, Dimension));
	void (* fill_elipse)
	  PR((MonetShade, Position, Position, Dimension, Dimension));
	void (* fill_rectangle)
	  PR((MonetShade, Position, Position, Dimension, Dimension));
	void (* fill_roundedRectangle)
	  PR((MonetShade, Position, Position, Dimension, Dimension));
	/* -- set_scale is really set text scale  */
	void (* set_scale)
	  PR((float, float));
	void (* set_reverse)	
	  PR((void));
	/* -- draw in reverse video, to emulate XOR drawing */
	MonetBool (* in_region)	
	  PR((Position, Position, Dimension, Dimension));
} * DrawLibrary;

#define NEWDRAWLIB(Lib, xor) \
	Lib = MonetNew (struct drawLibrary);   \
	Lib->end_draw = end_draw;               \
	Lib->draw_percentage_line = draw_percentage_line; \
	Lib->draw_boxed_string = draw_boxed_string; \
	Lib->draw_string = draw_string; \
	Lib->draw_line = draw_line;      \
	Lib->draw_elipse = draw_elipse;   \
	Lib->draw_rectangle = draw_rectangle;\
	Lib->draw_roundedRectangle = draw_roundedRectangle;\
	Lib->fill_elipse = fill_elipse;               \
	Lib->fill_rectangle = fill_rectangle;          \
	Lib->fill_roundedRectangle = fill_roundedRectangle;    \
	Lib->set_scale = set_scale; \
	Lib->set_reverse = set_reverse; \
        Lib->in_region   = in_region; \
        Lib->is_xor = xor;

#define create_drawLibrary()  (MonetNew (struct draw_library));
#define free_drawLibrary(l)    MonetFree (l);

/*********************************************************/
/*  Text  */
/*********************************************************/

extern MonetBool Draw_start_text PR ((String));

/*********************************************************/
/*  PS  */
/*********************************************************/

/*  returns filename  */
extern char * Draw_start_postscript
  PR ((String, String, String,
       String, String, String,
       Position, Position,
       Position, Position));

/*********************************************************/
/*  General  */
/*********************************************************/

extern void	MonetDraw_init
  PR ((void));
extern void	MonetDraw_finish
  PR ((void));

extern void Draw_start_draw
  PR ((DrawLibrary, Position x, Position y,  float x_scale, float y_scale));
extern void Draw_end_draw
  PR ((void));
extern void Draw_percentage_line
  PR ((MonetShade, unsigned, unsigned,
       Position, Position, Position, Position));

extern void Draw_string
  PR ((MonetShade, Position, Position,
       Dimension, Dimension, String,
       unsigned len));
extern void Draw_boxed_string
  PR ((MonetShade, Position, Position,
       Dimension, Dimension, String,
       unsigned len));

extern void  MonetDraw_adjust_for_scale
  PR((Position *x, Position *y, Dimension *width, Dimension *height));

extern void  MonetDraw_adjust_xy_for_scale
  PR((Position  *x, Position *y));

extern void  MonetDraw_adjust_wh_for_scale
  PR((Dimension *width, Dimension *height));

extern MonetBool Draw_significant_line
  PR ((MonetShade, Position, Position, Position, Position));
extern void Draw_line
  PR ((MonetShade, Position, Position, Position, Position));
extern void Draw_elipse
  PR ((MonetShade, Position, Position, Dimension, Dimension));
extern void Draw_circle
  PR ((MonetShade, Position, Position, Dimension));
extern void Draw_rectangle
  PR ((MonetShade, Position, Position, Dimension, Dimension));
extern void Draw_box
  PR ((MonetShade, MonetShade, Position, Position, Dimension, Dimension,
       Position));
extern void Draw_box_elipse
  PR ((MonetShade, MonetShade, Position, Position, Dimension, Dimension,
       Position));
extern void Draw_roundedRectangle
  PR ((MonetShade, Position, Position, Dimension, Dimension));
extern void Draw_fill_circle
  PR ((MonetShade, Position, Position, Dimension));
extern void Draw_fill_elipse
  PR ((MonetShade, Position, Position,Dimension,Dimension));
extern void Draw_fill_rectangle
  PR ((MonetShade, Position, Position,Dimension,Dimension));
extern void Draw_fill_roundedRectangle
  PR ((MonetShade, Position, Position, Dimension, Dimension));
extern MonetBool Draw_is_xor
  PR((void));
extern MonetBool Draw_in_region
  PR ((Position, Position, Dimension, Dimension));

extern MonetShade MonetShade_inverse
  PR((MonetShade));

extern void Draw_set_box
  PR ((Position x, Position y,
       Dimension in_w, Dimension in_h,
       Dimension out_w, Dimension out_h));
extern void Draw_reset_box
  PR ((void));
extern void Draw_set_reverse
  PR ((void));

#endif

