/*****************************************************************************/
/*****************************************************************************/
/* $Header$  */
/*****************************************************************************/
/* This deals with changes in network entity being reflected in its drawings */
/* their are two parts.  One for setting up exactly how a drawing will be    */
/* displayed given the current status of the entity it represents and a part */
/*  for handling changes as they occur.   These need to be rather generic.   */
/*****************************************************************************/

#include "mapP.h"
#include <monet/map/Lines.h>
#include <monet/db/Enum.h>
#include <monet/msh/Script.h>

typedef struct monet_mapping * Mapping;
typedef struct value_mapping * ValueMapping;

struct monet_mapping {
  MonetAttribute attribute;
  ValueMapping values;
  Mapping next;
};

struct value_mapping {
  unsigned value;
  MonetShapeState state;
  ValueMapping next;
};

static Mapping MappingList = NULL;
static MonetShapeState SelectMapping = 0;

static Drawing CurrentClass;

/*****************************************************************************/

/*****************************************************************************/

MONETCALLBACK(do_help_drawing_state)
     Drawing drawing;
     MonetMap map;
     if (map = MapCurrent_get ()) {
       MonetEntity e;
       e = String2MonetEntity (argv[1]);
       if (e) {
	 drawing = get_drawing (e, map);
	 if (drawing) {
	   MonetPrintfnl ("Drawing State of [%s] is [%lx]",
			   Drawing_title (drawing),
			   drawing->state);
	 }
       }
     }
MONETEND

MONETCALLBACK(do_help_draw_mapping)
     Mapping list;
     ValueMapping value;

     FORALL(list, MappingList, next) {
       MonetType type;
       MonetEnum en;
       MonetPrintfnl ("MonetMap attribute [%s]",
		       MonetAttribute2String (list->attribute));
       type = MonetAttribute2MonetType (list->attribute);
       en = MonetType2MonetEnum (type);
       FORALL(value, list->values, next) {
	 MonetPrintfnl ("  [%s] [%s]",
			 MonetEnum_unsigned2String (en, value->value),
			 MonetShapeState2String (value->state));
       }
     } 
MONETEND


/***************************************************************/
/*  Do this when the status of a drawing changes               */
/*  Checks all the attribute mapping before making a change.   */
/*  only if the attribute has a mapping is the drawing changed */
/***************************************************************/

static MonetBool check_change (num, entity)
     int num;
     MonetEntity entity;
{
  if (MonetMask_ISSET(num, &MASK(entity))) {
    Drawing d;
    FORALL(d, entity->drawing, next) {
      Drawing_changed_redraw (d, False);
    }
    return True;
  }
  return False;
}

static MonetBool check_attribute (map, drawing, attribute)
     MonetAttribute attribute;
     MonetMap map;
     Drawing drawing;
{
  if (attribute) {
    unsigned num;
    num = MonetAttribute2AttributeNumber (attribute);
    if (MonetMask_ISSET(num, &MASK(ENTITY(drawing)))) {
      Drawing_changed_redraw (drawing, False);
      return True;
    }
  }
  return False;
}

extern void Drawing_change_trigger (entity)
  MonetEntity entity;
{
  Drawing d;
  Mapping list;

  if (check_change (MonetAttribute_FROM_NUM, entity))
    return;
  if (check_change (MonetAttribute_TO_NUM, entity))
    return;
  if (check_change (MonetAttribute_SHAPE_NUM, entity))
    return;

  FORALL(list, MappingList, next) {
    int num;
    num = MonetAttribute2AttributeNumber (list->attribute);
    if (check_change (num, entity)) {
      return;
    }
  }
  FORALL(d, entity->drawing, next) {
    MonetMap map;

    map = Drawing_mapOf (d);

    if (check_attribute (map, d, map->primary_label))
      return;
    if (check_attribute (map, d, map->secondary_label))
      return;
    if (check_attribute (map, d, map->scale_from_line))
      return;
    if (check_attribute (map, d, map->scale_to_line))
      return;
  }
}

/***************************************************************/
/*   Set selection flag, and then redraw system */
/***************************************************************/

static void handle_select (drawing)
     Drawing drawing;
{
  switch (SELECT(drawing)) {
  default:
    MonetSysError("Invalid select state");
  case select_none:
    drawing->state = drawing->true_state;
    break;
  case select_normal:
  case select_super:
  case select_first:
    drawing->state = MonetShapeState_add (drawing->state, SelectMapping);
    break;
  }
  Prepare_connecting_lines (drawing);
}
/***************************************************************/
/***  For marking systems selected by the user ***/
/*    The poor naming is historical !!! */
/***************************************************************/

static void Drawing_select (drawing, select)
     Drawing drawing;
     SelectState select;
{
  if (SELECT(drawing) != select) {
    Drawing_unmoved_undraw (drawing);
    SELECT(drawing) = select;
    handle_select (drawing);
    Drawing_unmoved_redraw (drawing);
  }
}

static void Drawing_deSelect (drawing)
     Drawing drawing;
{
  Drawing_select (drawing, select_none);
}

static void    Drawing_select_no_draw (drawing, select)
     Drawing drawing;
     SelectState select;
{
  if (SELECT(drawing) != select) {
    SELECT(drawing) = select;
    handle_select (drawing);
  }
}

extern Drawing Drawing_current_class ()
{
  return CurrentClass;
}

extern void mark_class (o)
     Drawing o;
{
  /* Deselect previous selection */
  if (CurrentClass != NULL) {
    Drawing_deSelect (CurrentClass);
  }
  /*  Toggle chosen class */
  if (CurrentClass != o) {
    Drawing_select (o, select_normal);
    CurrentClass = o;
  }
}

extern void unmark_class ()
{
  Drawing_deSelect (CurrentClass);
  CurrentClass = NULL;
}

extern Drawing	current_Selection (map)
     MonetMap map;
{
  return (DrawingList_first (map->selected));
}

extern unsigned	numberOf_Selection (map)
     MonetMap map;
{
  return (DrawingList_sizeOf (map->selected));
}

extern Drawing	next_selection (map)
     MonetMap map;
{
  return (DrawingList_next (map->selected));
}

extern Drawing	second_selection (map)
     MonetMap map;
{
  return (DrawingList_second (map->selected));
}

extern void	add_selection (d)
  Drawing d;
{
  if (Drawing_isSelected (d)) {
    return;
  } else {
    MonetMap map;
    map = Drawing_mapOf (d);
    Drawing_select (d, select_normal);
    DrawingList_insert (map->selected, d);
  }
}

extern void	remove_selection (d)
  Drawing d;
{
  if (Drawing_isSelected (d)) {
    MonetMap map;
    map = Drawing_mapOf (d);
    Drawing_deSelect (d);
    DrawingList_extract (map->selected, d);
  }
}

extern void	toggle_selection (d)
     Drawing d;
{
  if (Drawing_isSelected (d))
    remove_selection (d);
  else
    add_selection (d);

}

extern void select_no_draw (d)
     Drawing d;
{
  if (Drawing_isSelected (d)) {
  } else {
    MonetMap map;
    map = Drawing_mapOf (d);
    Drawing_select_no_draw (d, select_normal);
    DrawingList_insert (map->selected, d);
  }
}

extern void free_selection (map)
     MonetMap map;
{
  Drawing s;

  FOR_DRAWINGLIST (s, map->selected) {
    Drawing_deSelect (s);
  } END_FORLIST;
  DrawingList_reset (map->selected);
}

extern MonetBool Drawing_isSelected (s)
	Drawing s;
{ return SELECT(s) != select_none; }

/************************************************************/
/* Terribly important routine, decides what a visual drawing */
/* looks like depending on its current status */
/* Prepare system for drawing */
/************************************************************************/

extern MonetBool Drawing_prepare (drawing)
     Drawing drawing;
{
  MonetMap    map;
  MonetEntity entity = ENTITY(drawing);
  Mapping mapping;
  MonetShapeState state = 0;

  map = Drawing_mapOf(drawing);
  if (map == NULL) {
    MonetSysError ("prepare: MonetMap is NULL");
    return False;
  }
  if (map->drawnas == drawnas_list)
    Drawing_reset_list_label (drawing);

  /************************************************************/
  /*  Set the label */
  /************************************************************/

  (void) Drawing_primary_label (drawing);

  (void) Drawing_secondary_label (drawing);

  /************************************************************/
  /*  Setup display state  */
  /************************************************************/

  FORALL (mapping, MappingList, next) {
    ValueMapping value;
    unsigned number, v;
    number = MonetAttribute2AttributeNumber (mapping->attribute);
    v = MONETGETINTEGER(entity, number);
    FORALL (value, mapping->values, next) {
      if (v == value->value) {
	state |= value->state;
      }
    }
  }
  drawing->shape      = MonetEntity_shapeOf (ENTITY(drawing));

  if (MonetShape_isNet (drawing->shape)) {
    switch (map -> net_border) {
    default:
    case netb_box:
      break;
    case netb_round:
      state |= MONETSTATE_ROUND_BORDER;
      break;
    case netb_minimal:
      state |= MONETSTATE_NO_BORDER;
      break;
    }
  }
  drawing->state      = state;
  drawing->true_state = state;

  MONETDEBUG(30, ( "[%s] drawing state", MonetEntity_title (entity), state));


  handle_select (drawing);

  /************************************************************/
  /*  Set up given the MonetShape                             */
  /*  Set up width and height of drawing if not set           */
  /************************************************************/

  MonetShape_prepare (drawing->shape,
		      Drawing_isClass (drawing),  /* Force a label if class */
		      map->connectors,
		      &WIDTH(drawing), &HEIGHT(drawing),
		      LABEL(drawing));

  /************************************************************/
  /*  Prepare the lines connected to the drawing  */
  /************************************************************/

  if (map->scale_from_line) {
    int number, v;
    number = MonetAttribute2AttributeNumber (map->scale_from_line);
    v = MONETGETINTEGER(entity, number);
    MonetConnectingLine_set (drawing->from, v, map->scale_from_by);
  } else {
    MonetConnectingLine_set (drawing->from, 0, 0);
  }
  if (map->scale_to_line) {
    int number, v;
    number = MonetAttribute2AttributeNumber (map->scale_to_line);
    v = MONETGETINTEGER(entity, number);
    MonetConnectingLine_set (drawing->to, v, map->scale_to_by);
  } else {
    MonetConnectingLine_set (drawing->to, 0, 0);
  }
  (void) Prepare_connecting_lines (drawing);

  return True;
}

/*****************************************************************************/
/*  free mapping list  */
/*****************************************************************************/

static void Mapping_free (list)
     Mapping list;
{
  while (list->values) {
    MonetFree (list->values);
    list -> values = list->values->next;
  }
  MonetFree (list);
  LIST_DEL (Mapping, list, MappingList, next);
}

/*****************************************************************************/
/*  Register drawnas attributes  */
/*****************************************************************************/

MONETCALLBACK (do_draw_mappings)
     Mapping mapping;
     MonetAttribute attribute;
     MonetEnum en;
     MonetType type;

     if (argc % 2 == 1) {
       MonetPrintfError ("draw-mapping: must have odd number of arguments");
     } else if (attribute = String2MonetAttribute (argv[1])) {
       type = MonetAttribute2MonetType (attribute);
       if (en = MonetType2MonetEnum (type)) {
	 if (mapping = MonetNew (struct monet_mapping)) {
	   mapping -> values = NULL;
	   LIST_ADD (mapping, MappingList, next);
	   mapping -> attribute = attribute;
	   while (argc > 2) {
	     ValueMapping newvalue;
	     String draw, value;
	     --argc;
	     draw  = argv[argc];
	     --argc;
	     value = argv[argc];
	     newvalue = MonetNew (struct value_mapping);
	     newvalue->value = MonetEnum_String2unsigned (en, value);
	     newvalue->state = String2MonetShapeState (draw);
	     LIST_ADD(newvalue, mapping->values, next);
	   }
	 } else {
	   MonetSysError ("Could not allocate space");
	 }
       } else {
	 MonetPrintfError ("[%s] is not an enumerated type atribute", argv[1]);
       }
     } else {
       MonetPrintfError ("No such attribute [%s]", argv[1]);
     }
MONETEND

MONETCALLBACK (do_select_mapping)
     SelectMapping = String2MonetShapeState (argv[1]);
MONETEND


MONETCALLBACK (do_clear_draw_mappings)
  while (MappingList)
    Mapping_free (MappingList);
MONETEND

/*****************************************************************************/
/*  Init and finish  */
/*****************************************************************************/

extern void monet_map_change_init ()
{
  MappingList = NULL;
  CurrentClass = NULL;
  (void) MonetMember_module_class ("map");

  MonetMember_create
    ("clear-draw-mapping", do_clear_draw_mappings,
     "",
     "Clear all mappings set using 'draw-mapping'");
  MonetMember_create
    ("draw-mapping", do_draw_mappings,
     "entity attribute  ?value ?drawnas ...",
     "Register a maping between an entity attribute what will be displayed. \
Eg. if reachable attribute then map this to an up flag");
  MonetMember_create
    ("select-mapping", do_select_mapping,
     "drawnas",
     "Register how a selected drawing is shown on the screen, \
use help-draw-mapping for a list of 'drawnas'");
  MonetMember_create
    ("help-draw-mapping", do_help_draw_mapping,
     "",
     "Lists mappings between entity attributes and drawing states");
  MonetMember_create
    ("help-drawing-state", do_help_drawing_state,
     "drawing",
     "Shows drawing state of drawing in hex");
}

/*****************************************************************************/
/*  Finish off */
/*****************************************************************************/

extern void monet_map_change_finish ()
{
  while (MappingList)
    Mapping_free (MappingList);
}

/*****************************************************************************/
/*****************************************************************************/
/* $Header$  */
/*****************************************************************************/
/*****************************************************************************/
