/******************************************************************/
/******************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/drawing.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/******************************************************************/
/*  Drawings are visual representations of network entities.      */
/*  Each drawing represents exactly one entity, but an entity may */
/*  be represented by more than one drawing.                      */
/*  A drawing appears on exactly one map.                         */
/*  All methods (routines) applicable to a drawing are prefixed   */
/*  by 'Drawing_'.                                                */
/*  A drawing could be implemented as an X11 widget or gadget.    */
/******************************************************************/

#include "mapP.h"
#include <monet/map/Interface.h>

static Drawing CurrentDrawing = NULL;

/*****************************************************************************/
/* Some debugging  */
/*****************************************************************************/

static int MallocedD = 0, FreedD = 0;

extern void Drawing_dump ()
{
  MonetPrintfnl ("sizeof a drawing [%d]", sizeof (struct monet_drawing));
  MonetPrintfnl ("Malloced : %d", MallocedD);
  MonetPrintfnl ("Freed    : %d", FreedD);
  MonetPrintfnl ("Diff     : %d", MallocedD - FreedD);
  MonetPrintfnl ("Space    : %d",
		 (MallocedD - FreedD) * sizeof (struct monet_drawing));
}

/*****************************************************************************/
/*  Current Drawing  */
/*****************************************************************************/

extern	Drawing Drawing_get_current ()
{
  return CurrentDrawing;
}

extern void Drawing_set_current (drawing)
     Drawing drawing;
{
  CurrentDrawing = drawing;
}

extern MonetBool Drawing_isMinimal (drawing)
     Drawing drawing;
{
  if (MonetShape_isNet (drawing->shape)) {
    if (Drawing_mapOf (drawing) -> net_border == netb_minimal) {
      return True;
    }
  }
  return False;
}

extern MonetBool Drawing_isElipse (drawing)
     Drawing drawing;
{
  if (MonetShape_isNet (drawing->shape)) {
    if (Drawing_mapOf (drawing) -> net_border == netb_round) {
      return True;
    }
  }
  return False;
}

/*****************************************************************************/
/*  Is a point outwith an elipse  */
/*****************************************************************************/

#define SQR(x) ((x)*(x))
static MonetBool outofElipseBounds (px,py, x,y, w,h)
     int px, py, x,y,h,w;
{
  if (px >= 0 && py >= 0) {
    int r;
    px -= x + w/2;
    py -= y + h/2;
    if (w > h) {
      py *= w;
      py /= h;
      r = w / 2;
    } else {
      py *= h;
      py /= w;
      r = h / 2;
    }
    if (SQR(px) + SQR(py) >  SQR(r)) {
      return True;
    }
  }
  return False;
}

/*****************************************************************************/
/*  Is a point outwith a rectangle  */
/*****************************************************************************/

static MonetBool outofBounds (x,y, xb,yb,wb,hb)
     Position x,y, xb,yb;
     Dimension hb,wb;
{
  if (x < 0 || y < 0)
    return False;
  if ((Dimension) x>=xb)
    if ((Dimension) y>=yb)
      if ((Dimension) x<=xb+wb)
	if ((Dimension) y<=yb+hb)
	  return False;
  return True;
}

/******************************************************************/
/*  Test if two drawings are linked together, for this to be true */
/*  one must be an interface and the oher an system  */
/******************************************************************/

extern MonetBool Drawing_are_linked (a, b)
     Drawing a,b;
{
  Drawing sys, interface;

  if (Drawing_isInterface (a)) {
    if (Drawing_isInterface (b))
      return False;
    interface = a;
    sys = b;
  } else if (Drawing_isInterface (b)) {
    interface = b;
    sys = a;
  } else
    return False;
  
  if (sys == Drawing_from_interface(interface) ||
      sys == Drawing_to_interface(interface))
    return True;

  return False;
}

/******************************************************************/
/* Test if a drawing can be resized */
/******************************************************************/

extern MonetBool Drawing_isResizable (o)
     Drawing o;
{
  if (Drawing_isClass (o))
    return False;
  if (MonetShape_resizable (MonetEntity_shapeOf (ENTITY(o))))
    return True;
  return False;
}

/******************************************************************/
/*  Select all of an entities drawings                            */
/******************************************************************/

extern void MonetEntity_select_drawings (e)
	MonetEntity e;
{
	Drawing d;
	FORALL (d, e->drawing, next)
		add_selection (d);
}

/******************************************************************/
/*  Allocate and init basic drawing struct                        */
/******************************************************************/

static Drawing basic_create ()
{
  Drawing drawing;

  drawing = MonetNew (struct monet_drawing);
  ++ MallocedD;
  drawing->select = select_none;
  drawing->map = NULL;
  drawing->label = NULL;
  drawing->given_label = NULL;
  drawing->resized = False;
  drawing->x = drawing->y = -1;  /*  Unplaced */
  drawing->stretch_height = drawing->stretch_width = 0;
  drawing->state = 0;
  drawing->true_state = 0;
  drawing->shape = NULL;
  drawing->label2 = NULL;
  drawing->labellen = 0;
  drawing->next = NULL;
  drawing->entity = NULL;
  drawing->redraw_connectors = True;
  drawing->list.x = 0;
  drawing->list.y = 0;
  drawing->list.width = 0;
  drawing->list.height = 0;
  drawing->list.label = NULL;

  drawing->primary_label = label_default;
  drawing->secondary_label = label_attribute;

  drawing->from = MonetConnectingLine_create ();
  drawing->to   = MonetConnectingLine_create ();

  drawing -> submap = NULL;
  drawing -> map_interfaces = NULL;
  drawing -> submap_visible = False;

  return drawing;
}

/*****************************************************************************/
/*  Returns the drawing of a class entity, or if it has none, creates one    */
/*  This is like Map_insert_drawing but for drawing classes                  */
/*  !!! This needs rationalised */
/*****************************************************************************/

extern Drawing Map_insert_drawing_class (map, entity_class)
     MonetEntity entity_class;
     MonetMap map;
{
  Drawing drawing;

  if (map == NULL) {
    MonetSysError ("insert: MonetMap is null");
    return NULL;
  }
  if (!Map_classesOnly(map)) {
    MonetSysError ("MonetMap [%s] is not reserved for classes", Map_title (map));
    return NULL;
  }
  if (entity_class == NULL)
    return NULL;

  if (entity_class->drawing == NULL) {
    if (!(drawing = basic_create ()))
      return NULL;
    drawing -> entity = entity_class;
    entity_class -> drawing = drawing;
    Drawing_mapOf(drawing) = map;
    DrawingList_insert(Map_elements (map), drawing);
    Drawing_prepare(drawing);
  }
  return entity_class->drawing;
}

/***************************************************************/
/*  Can an drawing be moved?                                    */
/***************************************************************/

extern MonetBool Drawing_isMoveable (o)
     Drawing o;
{
  return Drawing_isClass (o) == False;
}

/************************************************************************/
/* Find interfaces                                                      */
/* Will return a drawing on same map */
/************************************************************************/

extern Drawing Drawing_from_interface (d)
     Drawing d;
{
  MonetEntity from;

  if (from = from_interface (ENTITY(d))) {
    MonetMap map;
    Drawing result;
    map = Drawing_mapOf(d);
    FORALL (result, from->drawing, next)
      if (Drawing_mapOf(result) == map)
	return result;
  }
  return NULL;

}

extern Drawing Drawing_to_interface (d)
  Drawing d;
{
  MonetEntity to;

  if (to = to_interface (ENTITY(d))) {
    MonetMap map;
    Drawing result;
    map = Drawing_mapOf(d);
    FORALL (result, to->drawing, next)
      if (Drawing_mapOf(result) == map)
	return result;
  }
  return NULL;
}

/***************************************************************/
/*  Find a drawing given its entity and map                    */
/***************************************************************/

extern Drawing get_submap_drawing (entity, map)
     MonetEntity entity;
     MonetMap map;
{
  if (entity) {
    Drawing d;
    FORALL (d, entity->drawing, next) {
      MonetMap map_d;
      if (map_d = Drawing_mapOf(d)) {
	/* All drawings should have a map, but this is a hack since it
	   is called from, remove_connections, which is called from
	   extract_drawing.  Not very nice !!! */
	if (map_d != map) {
	  Drawing dm;
	  FORALL (dm, map_d->entity->drawing, next)
	    if (Drawing_mapOf(dm) == map) {
	    return dm;
	  }
	}
      }
    }
  }
  return NULL;
}	

/***************************************************************/
/*  Find a drawing given its entity and map                    */
/***************************************************************/

extern Drawing get_drawing (entity, map)
     MonetEntity entity;
     MonetMap map;
{
  if (entity) {
    Drawing d;
    FORALL (d, entity->drawing, next) {
      MonetMap map_d;
      map_d = Drawing_mapOf(d);
      if (map_d == map) {
	return d;
      }
    }
  }
  return NULL;
}	

/***************************************************************/
/*  Find a drawing given its name, class and map               */
/***************************************************************/

extern Drawing Drawing_find (title, entity_class, map)
     String title;
     MonetMap map;
     MonetEntity entity_class;
{
  MonetEntity entity;

  if (entity = String2MonetEntity_instance (entity_class, title))
    return get_drawing (entity, map);
  return NULL;
}

/***************************************************************/
/*  Create basic drawing */
/***************************************************************/

extern Drawing Drawing_create_from_entity (entity, map, shape,label,x, y, w, h)
     MonetEntity entity;
     MonetMap map;
     MonetShape shape;
     char * label;
     Position x, y;
     Dimension w,h;
{
  Drawing drawing;
  MonetEntity entity_class = MonetEntity_classOf(entity);

  if (map == NULL) {
    MonetSysError ("Map is Null, trying to create instance");
    return NULL;
  } else if (Map_can_insert_drawing (map, entity_class) == False) {
    char * class_title = MonetNewString (MonetEntity_title (entity_class));
    MonetPrintfError ("Cannot insert drawings of class [%s] in map [%s]",
		class_title, Drawing_title (map));
    MonetFree (class_title);
    return NULL;
  }
  if (drawing = basic_create ()) {
    RedirectMonetEntity_add (entity);
    drawing -> entity = entity;
    drawing -> next = entity -> drawing;
    entity -> drawing = drawing;
    if (MonetEntity_isMap (entity_class)) {
      MONETDEBUG(1, ("Creating drawing with map [%s]",
		  Drawing_title (drawing)));
      drawing -> submap = MapEntity_make (entity);
      drawing -> map_interfaces = DrawingList_create ();
    }
    Map_insert_drawing (map, drawing, shape,label,
			x, y, (Dimension) w, (Dimension) h);
    return drawing;
   }
  return NULL;
}

/***************************************************************/
/*  Create an instance */
/***************************************************************/

extern Drawing Drawing_create (title, entity_class, map, x, y, w, h)
	String title;
	MonetEntity entity_class;
	MonetMap map;
	Position x, y;
	Dimension w,h;
{
  MonetEntity entity;



  if (MonetEntity_isClass (entity_class) == False) {
    MonetSysError ("class is an instance");
    return NULL;
	}
  if (entity = String2MonetEntity_instance (entity_class, title)) {
    if (get_drawing (entity, map)) {
      MonetPrintfError ("Drawing [%s:%s] already exists on map [%s]",
		  MonetEntity_title (entity_class), title, Map_title (map));
      return NULL;
    }
  } else if (!(entity = MonetEntity_create_instance (entity_class, title)))
    return NULL;
 
  return Drawing_create_from_entity (entity, map, NULSHAPE, NULSTR, x,y, w,h);
}


/***************************************************************/
/*  Create an instance by full name */
/***************************************************************/

extern Drawing Drawing_create_by_name (name, map, x, y, w,h)
     char * name;
     MonetMap map;
     Position x;
     Position y;
     Dimension w;
     Dimension h;
{
  MonetEntity entity_class;
  String title;
  MonetBool isclass;
  MonetEntity entity;

  if (entity = parse_title (name, &title, &entity_class, &isclass))
    return Drawing_create_from_entity
      (entity, map, NULSHAPE, NULSTR, x,y, w,h);

  if (entity_class == NULL || title == NULL) {
    MonetPrintfError("Can't parse title [%s]", name);
    return NULL;
  } else if (isclass) {
    MonetPrintfError("Can't create drawing of class [%s]", name);
    return NULL;
  }
  return  Drawing_create (title, entity_class, map, x, y,w,h);
}

/******************************************************************/
/*  Returns true if any part of drawing 'o' is in box [x,y][x+w,y+h] */
/******************************************************************/

#define overlap(x1,w1,x2,w2) ((x1+w1 > x2) && (x1 < x2+w2))
extern MonetBool Drawing_in_box (o,x,y,w,h)
	Drawing o;
	Position x,y;
	Dimension w,h;
{
	Dimension pw, ph;
	Position  px, py;
	MonetMap map = Drawing_mapOf(o);

	switch (map->drawnas) {
	case drawnas_list:
		px = LISTBOX(o).x;
		py = LISTBOX(o).y;
		pw = LISTBOX(o).width;
		ph = LISTBOX(o).height;
		break;
	case drawnas_map:
		px = X(o);
		py = Y(o);
		pw = WIDTH(o);
		ph = HEIGHT(o);
		break;
	default:
		MonetSysError ("Invalid drawnas[%d], [%s]", map->drawnas,
				Drawing_title(o));
		return False;
	}
	/*  True if there is an overlap in the x and the y  */
	return (overlap (x,w, px, pw) && overlap (y,h, py,ph));
}


/*  Returns true if [x,y] correctly mark objetc [o] */
extern MonetBool Drawing_is_marked (o,px,py)
     Drawing o;
     Position px,py;
{
  MonetMap map = Drawing_mapOf(o);

  switch (map->drawnas) {
  case drawnas_list:
    return (!outofBounds (px,py, LISTBOX(o).x, LISTBOX(o).y,
			  LISTBOX(o).width, LISTBOX(o).height));
  case drawnas_map:
    if (Drawing_isElipse (o)) {
      return (!outofElipseBounds (px,py, X(o), Y(o), WIDTH(o), HEIGHT(o)));
    } else {
      return (!outofBounds (px,py, X(o), Y(o), WIDTH(o), HEIGHT(o)));
    }
  default:
    MonetSysError ("Invalid drawnas[%d], [%s]", map->drawnas,
		   Drawing_title(o));
    return False;
  }
  /* NOTREACHED */
}

/*****************************************************************************/
/* Test to see if an drawing is a map */
/*****************************************************************************/

extern MonetBool Drawing_isMap(s)
     Drawing          s;
{
  return MonetEntity_isMap (ENTITY(s));
}

/***************************************************************/
/*  Free drawing                                               */
/***************************************************************/

extern void Drawing_free (d)
     Drawing d;
{
  MonetEntity entity;
  if (d == Drawing_current_class ()) {
    unmark_class ();
  }
  remove_selection (d);

  Map_extract_drawing (Drawing_mapOf(d), d);
  if (d->label2 != NULL)
    MonetFree (d->label2);
  if (d->list.label != NULL)
    MonetFree (d->list.label);
  if (entity = ENTITY(d)) {
    /*  Delete from list of entities  */
    LIST_DEL(Drawing, d, entity->drawing, next);
    d->entity = NULL;
    RedirectMonetEntity_delete (entity);
  }
  MonetConnectingLine_free (d);
  if (d -> map_interfaces) {
    /*  This is a reference to a map  */
    DrawingList_free (d->map_interfaces);
  }
  if (d == CurrentDrawing)
    CurrentDrawing = NULL;
  ++FreedD;
  MonetFree (d);
}

/******************************************************************/
/*  Init and finish drawing module                                */
/******************************************************************/

extern void MonetDrawing_init ()
{
  CurrentDrawing = NULL;
}

extern void MonetDrawing_finish ()
{
  CurrentDrawing = NULL;
  /* finish with standard classes */
}

/***************************************************************/
/***************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/drawing.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/***************************************************************/
/***************************************************************/

