/***************************************************************/
/***************************************************************/
/* $header$ */
/***************************************************************/
/*  This is the source for the core of the MonetEntity stuff   */
/*  creation and deletion etc.                                 */
/***************************************************************/

#include <monet/msh/Entity.h>
#include <monet/msh/StringTable.h>
#include <monet/msh/Script.h>
#include "scriptP.h"
#include "monetP.h"

MonetAttribute  MonetClassAttribute = NULL; /* Every entity has a class */

/***************************************************************/
/*  Globals  */
/***************************************************************/

MonetEntity Entities; /* List of all entities  */
static MonetEntityList Cursors;
static MonetEntity MonetTopClass;

/*****************************************************************************/
/*  The class structure, internal, private class data                        */
/*  We seem to have three types of class, this is not right is it            */
/*  Some attributes are inherited, some aren't need to clarify this          */
/*  This structure contains special class stuff, plus attributes that        */
/*  cann only  be inherited                                                  */
/*****************************************************************************/

struct class_data {
  MonetType type;
  MonetStringTable instance_names;
  MonetEntity next;
  MonetEntityList      instances, subclasses;
};

static MonetEntity ClassList = NULL;

#define CLASS(obj)     (obj -> class_stuff)
#define INSTANCES(c)   (CLASS(c)->instances)
#define SUBCLASSES(c)   (CLASS(c)->subclasses)

/***************************************************************/
/*  Debgging stuff  */
/***************************************************************/

static int MallocedE = 0, FreedE = 0;

extern void MonetEntity_dump ()
{
  MonetPrintfnl ("sizeof an entity [%d]", sizeof (struct monet_entity));
  MonetPrintfnl ("Malloced : %d", MallocedE);
  MonetPrintfnl ("Freed    : %d", FreedE);
  MonetPrintfnl ("Diff     : %d", MallocedE - FreedE);
  MonetPrintfnl ("Space    : %d",
		 (MallocedE - FreedE) * sizeof (struct monet_entity));
}

/***************************************************************/
/*  Top class */
/***************************************************************/

extern MonetEntity MonetEntityTopClass ()
{
  return MonetTopClass;
}

/******************************************************************/
/*  Find an instance given its class and name                     */
/******************************************************************/

extern MonetEntity String2MonetEntity_instance (entity_class, name)
     char * name;
     MonetEntity entity_class;
{
  if (entity_class == NULL) {
    MonetSysError ("String2MonetEntity_instance : entity_class is null [%s]", name);
    return NULL;
  }
  return (MonetEntity)
    MonetStringTable_lookup (CLASS(entity_class)->instance_names, name);
}

/****************************************************************/
/*  Return the instances of a class                             */
/****************************************************************/

extern MonetEntityList MonetEntity_instances (entity_class)
	MonetEntity entity_class;
{
  return INSTANCES(entity_class);
}

extern MonetEntityList MonetEntity_subclasses (entity_class)
	MonetEntity entity_class;
{
  return SUBCLASSES(entity_class);
}

/***************************************************************/
/*  Create a new entity                                        */
/***************************************************************/

static MonetEntity basic_create (entity, entity_class, title)
     MonetEntity entity;
     char * title;
     MonetEntity entity_class;
{
  if (title == NULL) {
    MonetSysError ("NULL title");
    return NULL;
  }
  entity->class_stuff	= NULL;
  entity -> event          = NULL;
  entity -> redirect	= NULL;
  entity -> drawing	= NULL;
  entity -> map		= NULL;
  entity -> members         = NULL;
  entity->interfaces	= NULL;
  entity -> requests	= NULL;
  entity -> subcomponents = NULL;
  entity -> supercomponents = NULL;
  entity -> stacks = NULL;
  MonetStatusSet_create (entity);
  MONETTITLE(entity)	= MonetStr_refer (title);
  CLASSOFENTITY(entity)    = (caddr_t) entity_class;
  LIST_ADD (entity, Entities, next);

  return (entity);
}

/***************************************************************/
/*  Create an MonetEntity Instance                                  */
/***************************************************************/

extern MonetEntity MonetEntity_create_instance (entity_class, title)
	char * title;
	MonetEntity entity_class;
{
  MonetEntity entity = NULL;

  if (entity_class == NULL) {
    MonetSysError ("NULL class when creating entity");
  } else if (!MonetEntity_isClass (entity_class)) {
    MonetSysError ("Not a class");
  } else {
    String p;
    for (p = title;*p != NULL; p++)
      if (!isgraph (*p)) {
	MONETDEBUG(20, ( "Invalid title [%s] for entity", title));
	return NULL;
      }
    if (entity = MonetNew (struct monet_entity)) {
      ++ MallocedE;
      if (MonetStringTable_insert_unique
	  (CLASS(entity_class)->instance_names, title, (caddr_t) entity)) {
	MonetPrintfError ("MonetEntity already exists [%s:%s]",
		    MonetEntity_title (entity_class),title);
	MonetFree (entity);
	++ FreedE;
	return NULL;
      } else if (basic_create (entity, entity_class, title)) {
	MonetEntityList_insert (INSTANCES(entity_class), entity);
      } else {
	MonetFree (entity);
	++ FreedE;
	return NULL;
      }
    }
  }
  return entity;
}

/******************************************************************/
/*  Stuff specific to classes  */
/******************************************************************/

static void set_class_defaults (entity_class, type)
     MonetEntity entity_class;
     MonetType type;
{
  CLASS(entity_class)			= MonetNew (struct class_data);
  CLASS(entity_class) -> instances	= MonetEntityList_create ();
  /* !!! Arbitrary hash table size but no danger */
  CLASS(entity_class) -> instance_names= MonetStringTable_create (200);
  LIST_ADD (entity_class, ClassList, class_stuff->next);
  CLASS(entity_class) -> type          = type;
  CLASS(entity_class) -> subclasses	= MonetEntityList_create ();
}

/***************************************************************/
/*  Accessing user defined class types  */
/***************************************************************/

#define NEQ_NEW_CLASS(c,v) (c != v)
#define SET_NEW_CLASS(c,v) { c = v; }
#define GET_NEW_CLASS(o) (MONETGETPOINTER(o, id))
ACCESS(access_new_class, GET_NEW_CLASS,
       MonetEntity_title, String2MonetEntity,
       MonetEntity_title, String2MonetEntity,
       MonetEntity, NEQ_NEW_CLASS, SET_NEW_CLASS, 0);

/***************************************************************/
/*  Find a class given its name  */
/***************************************************************/

extern MonetEntity String2MonetClass (name)
     String name;
{
  return (MonetEntity) MonetType_find_data (name, access_new_class);
}

/***************************************************************/
/*  Each entity has a class, this type defines that class */
/***************************************************************/

static String access_class (o, id, op, format, null_default, data)
     MonetEntity o;
     unsigned id;
     AttributeOperation op;
     AttributeFormat format;
     MonetBool null_default;
     String data;
{
  MonetEntity *entity_class;

  if ( ! o)
    return NULL;
  entity_class = (MonetEntity * ) &(CLASSOFENTITY(o));
  if (op == attribute_init) {
    *entity_class = NULL;
  } else if (op == attribute_free) {
    *entity_class = NULL;
  } else if (op == attribute_unset) {
  } else if (op == attribute_set) {
    return NULL;
  } else if (op == attribute_get_shallow) {
  } else if (op == attribute_get) {
  } else {
    MonetSysError ("Unknown attribute access operation [%d]", (int) op);
  }
  if (*entity_class)
    return MonetEntity_title (*entity_class);
  return NULL;
}

/***************************************************************/
/*  Create an MonetEntity Class                                     */
/***************************************************************/

extern MonetEntity MonetEntity_create_class (entity_class, title)
     char * title;
     MonetEntity entity_class;
{
  MonetEntity entity;
  String p;
  MonetType type;

  if (entity_class == NULL) {
    entity_class = MonetTopClass;
  }
  for (p = title;*p != NULL; p++)
    if (*p == ':' || !isgraph (*p)) {
      MonetPrintfError ("Invalid title [%s] for entity_class", title);
      return NULL;
    }
  if (String2MonetEntity_subclass (entity_class, title) != NULL) {
    MonetPrintfError ("[%s] already exists", title);
    return NULL;
  }
  if (entity = MonetNew (struct monet_entity)) {
    ++ MallocedE;
    if (type = MonetType_create (title, False, access_new_class,
				 (caddr_t) entity, "A class type")) {
      if (basic_create (entity, entity_class, title)) {
	set_class_defaults (entity, type);
	MonetEntityList_insert (SUBCLASSES (entity_class), entity);
	return entity;
      } else {
	MonetType_free (type);
	MonetFree (entity);
	++ FreedE;
      }
    } else {
      MonetFree (entity);
      ++ FreedE;
    }
  }
  return NULL;
}

/***************************************************************/
/*  Free an entity                                             */
/*  This is very difficult to get right in its current form,   */
/*  A more automatic way would be to either count references   */
/*  and don't do any freeing until they are all accounted for. */
/*  Or only allow explicit two way links between things (hmm). */
/***************************************************************/

extern void MonetEntity_free (entity)
     MonetEntity entity;
{
  MonetEntity entity_class;

  entity_class = MonetEntity_classOf(entity);
  if (MonetEntity_isClass (entity)) {
    if (MonetEntityList_isEmpty (INSTANCES (entity)) == False) {
      MonetPrintfError ("Cannot delete class with instances");
    } else if (MonetEntityList_isEmpty (SUBCLASSES (entity)) == False) {
      MonetPrintfError ("Cannot delete class with subclasses");
    } else {
      if (entity_class) {
	MonetEntityList_extract (SUBCLASSES (entity_class), entity);
      }
      MonetStringTable_free (CLASS(entity)->instance_names);
      MonetEntityList_free (INSTANCES(entity));
      MonetEntityList_free (SUBCLASSES(entity));
      MonetType_free (CLASS(entity)->type);
      LIST_DEL (MonetEntity, entity, ClassList, class_stuff->next);
      MonetFree (CLASS(entity));
    }
    /* return; */
  } else if (entity_class == NULL) {
    MonetSysError ("MonetEntity's class is NULL");
  } else {
    MonetEntityList_extract (INSTANCES(MonetEntity_classOf(entity)), entity);
    MonetStringTable_remove (CLASS(entity_class)->instance_names,
			     TITLE (entity), (caddr_t) entity);
  }

  MonetEntity_free_members (entity);
  MonetStatusSet_free (entity);

  LIST_DEL (MonetEntity, entity, Entities, next);

  MonetEvent_free_entity (entity);
  MonetStack_free_entity (entity);

  MonetFree (entity);
  ++ FreedE;
}

/***************************************************************/
/*  Init and Finish                                            */
/***************************************************************/

extern void MonetEntity_init ()
{
  MonetEntity top;
  MonetType toptype, entity_class;

  Entities = NULL;
  monet_entity_stack_init ();
  Cursors = MonetEntityList_create ();
  MonetType_init ();

  /*********************************************/
  /*  Defaults types and attributes            */
  /*********************************************/

  if (top = MonetNew (struct monet_entity)) {
    ++ MallocedE;
    if (!basic_create (top, (MonetEntity) NULL, "Top")) {
      MonetFree (top);
      ++ FreedE;
      MonetSysError ("Could not initialise top class");
    }
  } else {
    MonetSysError ("Could not create top class");
  }
  toptype = MonetType_create (TITLE(top), False, access_new_class,
			   (caddr_t) top, "Top of the class heirarchy");
  set_class_defaults (top, toptype);
  MonetTopClass = top;

  entity_class = MonetType_create
    ("Class", True, access_class, NULL,
     "Each entity has a superclass, apart from Top");
  MonetClassAttribute =  MonetAttribute_create
    ("classof", MonetAttribute_CLASS_NUM, entity_class,
     "for classes only");
}

extern void MonetEntity_finish ()
{
  MonetType_finish ();
  MonetEntityList_free (Cursors);
  monet_entity_stack_finish ();
}

/***************************************************************/
/***************************************************************/
/* $Header$ */
/***************************************************************/
/***************************************************************/
