
/**********************************************************/
/**********************************************************/
/* $Header$ */
/**********************************************************/
/*  Entities have a list of pending events (MonetEvent).  */
/*  Each of these has a delay before handling and when    */
/*  ready a priority for handling.                        */
/**********************************************************/

#include <monet/msh/Entity.h>
#include <monet/msh/Event.h>
#include <monet/msh/Script.h>
#include <monet/msh/ScriptTree.h>

/**********************************************************/
/*  Definitions  */
/**********************************************************/

/**********************************************************/
/*  Implementation consists of a lists of delays.  Each   */
/*  'delay' has the absolute time it becomes pending and  */
/*  a list of events that become pending at that time.    */
/*  Each entity also has a list of events that are pending*/
/*  on it, so hen the entity is destroyed so are all of   */
/*  the events.                                           */
/*  The 'delay' lists are fifo queues to avoid starvation.*/
/**********************************************************/

typedef struct event_list * DelayList;
struct event_list {
  long when;
  MonetEvent events, last_event;
  DelayList next;
};

/***************************************************************************/
/* The event structure */
/***************************************************************************/

struct monet_event {
  MonetEvent next, member_next, entity_next;
  MonetMember member;
  MonetEntity entity;
  DelayList delay;
};

static DelayList MonetDelayList = NULL;
unsigned NumSpares = 0;

/**********************************************************/
/*  Represents an event as a string  */
/*  More for debugging than anything else  */
/**********************************************************/

extern String MonetEvent2String (event)
     MonetEvent event;
{
  static char buffer [1024]; /* !!! */
  (void) sprintf (buffer, "Event [%s][%s]",
		  MonetEntity_title (event->entity),
		  MonetMember2String (event->member));
  return buffer;
}

/**********************************************************/
/*  Are there any events ?  */
/**********************************************************/

extern MonetBool MonetEvent_anythere ()
{
  return MonetDelayList != NULL;
}

/**********************************************************/
/* -- Handle the next event and return the time when the  */
/* -- Looks at the setting of timer.tv_sec and if this is */
/* -- greater than the time when the next event becomes   */
/* -- pending, modifies timer accordinly, otherwise timer */
/* -- is not changed.                                     */
/**********************************************************/

extern int MonetEvent_next (timer)
     struct timeval *timer;
{
  if (MonetDelayList) {
    long delay;

    delay = MonetDelayList->when - time (NULL);

    if (delay <= 0) {
      MonetEvent event = MonetDelayList->events;
      MonetMember member = event -> member;
      MonetEntity entity = event->entity;

      MONETDEBUG(10, ("Doing event [%s]", MonetEvent2String (event)));
      MonetEvent_free (event);
      (void) MonetMember_execute (entity, member, NULL, NULL, NULL);
    }

    if (MonetDelayList) {
      delay = MonetDelayList->when - time (NULL);
      if (delay < 0) {
	delay = 0;
      }
      if (delay < timer->tv_sec) {
	timer->tv_sec = delay;
      }
      MONETDEBUG(20, ("Timer %ld, when %ld", timer->tv_sec, delay));
      return MonetDelayList->when;
    }
  }
  return -1;
}

/**********************************************************/
/*  Display state of all events  */
/**********************************************************/

MONETCALLBACK(do_help_events)
     MonetEvent event;
     unsigned count = 0;
     DelayList delay;

     MonetPrintfnl ("Current Event List");
     FORALL (delay, MonetDelayList, next) {
       MonetBool first = True;
       MonetPrintf ("Delay %lu %ld", delay->when, delay->when - time(NULL));
       FORALL (event, delay->events, next) {
	 if (first)
	   first = False;
	 else
	   MonetPrintf ("\t\t");
	 MonetPrintfnl ("\t%s", MonetEvent2String (event));
	 ++ count;
       }
     }
     if (count == 0) {
       MonetPrintf ("No events");
     } else if (count == 1) {
       MonetPrintf ("One event");
     } else {
       MonetPrintf ("%lu events", (long) count);
     }
     MonetPrintfnl ("");
MONETEND

/**********************************************************/
/*  Create new delay */
/**********************************************************/

static DelayList Delay_create (next, first_event, when)
     DelayList next;
     MonetEvent first_event;
     long when;
{
  DelayList delay;
  delay = MonetNew (struct event_list);
  delay -> when  = when;
  delay -> events = first_event;
  delay -> last_event = first_event;
  first_event->delay = delay;
  delay -> next = next;
  return delay;
}
/**********************************************************/
/*  Create new event  */
/**********************************************************/

extern MonetEvent MonetEvent_create (entity, member, delay)
     MonetEntity entity;
     MonetMember member;
     unsigned delay;
{
  MonetEvent event;
  DelayList list;
  long when;

  if (member == NULL)
    return NULL;

  when = delay + time(NULL);

  event = MonetNew (struct monet_event);
  if (event == NULL) {
    MonetSysError ("Could not allocate space for new MonetEvent");
  } else {
    MonetEvent event_list;
    event -> member = member;
    event -> next   = NULL;
    event -> entity = entity;
    LIST_ADD (event, entity->event, entity_next);
    event_list = MonetMember_event (member);
    LIST_ADD (event, event_list, member_next);
    if (MonetDelayList == NULL) {
      MonetDelayList = Delay_create (NULL, event, when);
    } else if (MonetDelayList->when >= when) {
      MonetDelayList = Delay_create (MonetDelayList, event, when);
    } else {
      FORALL (list, MonetDelayList, next) {
	if (when == list->when) {
	  list->last_event->next = event;
	  list->last_event = event;
	  break;
	} else if (list->next == NULL || list->next->when >= when) {
	  list->next = Delay_create (list->next, event, when);
	  break;
	}
      }
    }
  }
  return event;
}

/**********************************************************/
/*  Remove an entry from the delaylist  */
/**********************************************************/

static void DelayList_free (list)
     DelayList list;
{
  while (list && list->events)
    MonetEvent_free (list->events);
}

/**********************************************************/
/*  Free all the events belonging to an entity by name */
/**********************************************************/

extern void MonetEvent_free_entity_by_name (entity, name)
     MonetEntity entity;
     MonetStr name;
{
  MonetEvent event;
  for (event =  entity->event; event;) {
    if (MonetMember2Str (event->member) == name) {
      MonetEvent next;
      next = event->entity_next;
      MonetEvent_free (event);
      event = next;
    } else {
      event = event->entity_next;
    }
  }
}

/**********************************************************/
/*  Free a event  */
/**********************************************************/

extern void MonetEvent_free (event)
     MonetEvent event;
{
  MonetEvent list;
  LIST_DEL (MonetEvent, event, event->entity->event, entity_next);
  list = MonetMember_event (event->member);
  LIST_DEL (MonetEvent, event, list, member_next);
  if (event == event->delay->events) {
    event->delay->events = event->next;
    if (event->delay->events == NULL) {
      LIST_DEL (DelayList, event->delay, MonetDelayList, next);
      MonetFree (event->delay);
    }
  } else {
    MonetEvent e;
    FORALL (e, event->delay->events, next) {
      if (e->next == event) {
	e->next = event->next;
	if (event->delay->last_event == event) {
	  event->delay->last_event = e;
	}
	break;
      }
    }
  }
  MonetFree (event);
}

/***************************************************************/
/*  Do unique action for entity  */
/***************************************************************/

extern MonetBool MonetEntity_do_action (entity, event_name, delay)
     MonetEntity entity;
     String event_name;
     unsigned delay;
{
  MonetMember member;
  MonetStr id;
  MonetEntity qualifier;
  MonetEvent e;

  if ((id = String2MonetStr (event_name)) &&
      (member = MonetEntity_find_member (entity, id, &qualifier))) {
    FORALL(e, entity->event, entity_next)
      if (e -> member == member)
	MonetEvent_free (e);
      (void) MonetEvent_create (qualifier, member, delay);
    return True;
  }
  MonetPrintfError ("No such function [%s] for [%s]",
		    event_name, MonetEntity_title (entity));
  return False;
}

/*****************************************************************************/
/*  MonetEntity actions/triggers  */
/*****************************************************************************/

MONETCALLBACK(do_action_after)
     /*  !!! Make sure nothing is already running, no recursive
	 !!! action calls?  */
     (void) MonetEntity_do_action (entity, argv[2], String2unsigned (argv[1]));
MONETEND

MONETCALLBACK(do_clear_events)
  while (MonetDelayList) {
    DelayList_free (MonetDelayList);
  }
MONETEND

MONETCALLBACK(do_remove_event)
     MonetEvent_free_entity_by_name (entity, String2MonetStr(argv[1]));
MONETEND

/****************************************************************************/
/*  Init and finish  */
/****************************************************************************/

extern void MonetEvent_init ()
{
  MonetDelayList = NULL;

  (void) MonetMember_module_class ("event");
  MonetMember_create ("help",  do_help_events,
     "",
     "Displays the event list");
  MonetMember_create ("clear",  do_clear_events,
     "",
     "Clear all pending events");

  MonetMember_set_entity (MonetEntityTopClass ());
  MonetMember_create ("after", do_action_after,
     "delay function",
     "Perform an action, specific to the entity, after an optional delay. \
Even if delay is zero, the event will only be put on the pending list.");
  MonetMember_create ("remove-event", do_remove_event,
     "member",
     "Removes event from queue");

}

extern void MonetEvent_finish ()
{
  while (MonetDelayList) {
    DelayList_free (MonetDelayList);
  }
}

/**********************************************************/
/**********************************************************/
/* $Header$ */
/**********************************************************/
/**********************************************************/
