/***************************************************************/
/***************************************************************/
/* $header$ */
/***************************************************************/
/*  Given a string finds the entity and produces unique names  */
/*  given an entity.  In addition provides routines for finding*/
/*  and check subclasses.  */
/*  Basically a set of tools that don't depend on the innards  */
/*  of the MonetEntity structure and are not central to it.    */
/***************************************************************/

#include <monet/msh/Entity.h>
#include <monet/msh/StringTable.h>
#include <monet/msh/Script.h>

/******************************************************************/
/*  Find an entity class given its superclass and name            */
/******************************************************************/

extern MonetEntity String2MonetEntity_subclass (superclass, name)
     char * name;
     MonetEntity superclass;
{
  MonetEntity o;
  MonetStr str;

  if (superclass == NULL)
    return NULL;
  if (str = String2MonetStr (name)) {
    FOR_ENTITYLIST (o, MonetEntity_subclasses (superclass))
      if (MONETTITLE(o) == str)
	return o;
  } END_FORLIST;
  return NULL;
}

/******************************************************************/
/*  Is one class a subclass of another?                           */
/******************************************************************/

extern MonetBool MonetEntity_subclass (of, sub)
     MonetEntity of, sub;
{
  MonetEntity parent;

  if (of == NULL)
    return True;
  for (parent=sub; parent; parent = MonetEntity_classOf(parent))
    if (parent==of) 
      return True;
  return False;
}

/******************************************************************/
/*  Find an entity class given its full name                      */
/******************************************************************/

extern MonetEntity String2MonetEntityClass (s)
     String s;
{
  MonetBool isclass;
  MonetEntity superclass, result;
  char *name;
	
  result = parse_title (s, &name, &superclass, &isclass);
  if (isclass && name == NULL)
    return result;

  return (NULL);
}

/******************************************************************/
/*  Get the instance and class parts of an entity name  */
/******************************************************************/

extern void MonetEntity_class_instance_parts (string, class_name, instance)
     String string, *instance, *class_name;
{
  String s, c;
  static char buf[256]; /* !!! */
  c = buf;
  *instance = NULL;
  for (s = string; *s; ++s)
    if (*s == ':') {
      if (*(s+1))
	*instance = s+1;
      else
	*instance = NULL;
      break;
    } else if (c-buf + 1 < sizeof buf) {
      *c = *s;
      ++c;
    }
  *c = NULL;
  *class_name = buf;
}

/******************************************************************/
/*  Return unique title of an entity */
/*  Stores name in private memory, one of two buffers, this is so */
/*  it can be called twice in a printf statement without overwriting */
/*  the buffer area.  */
/******************************************************************/

extern char * MonetEntity_title (entity)
     MonetEntity entity;
{
  if (entity == NULL) {
    MonetSysError ("entity_title: MonetEntity is NULL");
  } else if (MonetEntity_isClass (entity)) {
    return TITLE (entity);
  } else {
    static char  buf1 [2048], buf2 [2048]; /* !!! */
    static String buffer = buf1;
    if (buffer == buf1)
      buffer = buf2;
    else
      buffer = buf1;
    (void) sprintf (buffer, "%s:%s",
		    TITLE(MonetEntity_classOf (entity)), TITLE(entity));
    return buffer;
  }
  return NULL;
}

/***************************************************************/
/*  Search for an entity by its name  */
/*  !!! Assumes unique names, must be changed for this reason  */
/***************************************************************/

extern MonetEntity String2MonetEntity (title)
     char * title;
{
  MonetBool isclass;
  MonetEntity superclass, result;
  char *name;

  result = parse_title (title, &name, &superclass, &isclass);
  if (name == NULL)
    return result;
  return (NULL);
}


/***************************************************************/
/*  Sets the official name of a system */
/***************************************************************/

extern MonetBool MonetEntity_rename (entity, name)
     String name;
     MonetEntity entity;
{
/*  if (string2instance (MonetEntity_classOf(entity), name)) {
    MonetPrintfInform ("Can't rename [%s] to [%s] as it already exists",
		 MonetEntity_title(entity), name);
    return False;
  } else if ( ! MonetEntity_is_title_valid (MonetEntity_classOf(entity),name)){
    MonetPrintfError (
		"[%s] is an invalid label for instances of class [%s]",
		name, MonetEntity_title (entity));
    return False;
    } else if (name && * name) {
    MonetStr_derefer (MONETTITLE(entity));
  MONETTITLE(entity) = MonetStr_refer (name);
    FORALL(d, entity->drawing, next)
      Drawing_redrawreset_label (d);
  } else {
    MonetPrintfError ("Name is null");
  }
*/
  MonetPrintfWarning ("rename disabled, obsolete");
  return True;
}

/******************************************************************/
/* Returns entity if found */
/* If part of subclass is not found, new is NULL otherwise it has  \*/
/* the name of the last piece */
/******************************************************************/

extern MonetEntity parse_title (title, name, superclass, isclass)
     char * title, **name;
     MonetEntity	* superclass;
     MonetBool	* isclass;
{
  MonetEntity result;
  char class_name [1024], *p;
  unsigned class_pointer = 0;

  if ( ! title || (* title == NULL)) {
    *superclass = NULL;
    *isclass = False;
    *name = NULL;
    return NULL;
  }
  for (p=title; *p; ++p, ++class_pointer) {
    if (class_pointer >= sizeof (class_name)) {
      MonetPrintfInform ("Class title is too long");
      *isclass = True;
      return NULL;
    }
    if (*p == ':') {
      *name = p+1;
      *isclass = False;
      class_name[class_pointer] = '\0';
      if (*name == '\0') {
	MonetPrintfError
	  ("MonetEntity title is <class>:<name>, name is null [%s]", title);
      } else if (*superclass = String2MonetClass (class_name)) {
	if (result = String2MonetEntity_instance (*superclass, *name)) {
	  *name = NULL;
	}
	return result;
      } else {
	MonetPrintfError ("Don't know class [%s]", class_name);
      }
      *name = NULL;
      return NULL;
    }
    class_name[class_pointer] = *p;
  }
  class_name[class_pointer] = '\0';
  *isclass = True;
  if (result = String2MonetClass (class_name)) {
    *name = NULL;
    *superclass = MonetEntity_classOf (result);
  } else {
    *name = class_name;
    *superclass = NULL;
  }
  return result;
}

/***************************************************************/
/***************************************************************/
/* $Header$ */
/***************************************************************/
/***************************************************************/

