/*****************************************************************************/
/*****************************************************************************/
/* $Header: script_parse.c,v 1.1 91/09/20 10:11:59 a38 Exp $ */
/*****************************************************************************/
/*****************************************************************************/

#include "scriptP.h"

static MonetTree Current, Source, Debris;
static MonetBool Error;

/*****************************************************************************/
/*  Dispose of the parse tree  */
/*****************************************************************************/

static MonetBool InExpression = False;  /* Parsing expression or statement */
static MonetTree parse_op PR ((MonetLex));

static MonetTree parse_args PR ((void));
static MonetTree parse_simple_term PR ((void));
static MonetTree parse_arg PR ((void));
static MonetTree parse_statement PR ((void));
static MonetTree parse_program PR ((void));
static MonetTree parse_instruction PR ((void));
static MonetTree parse_and_or PR ((void));

/*****************************************************************************/
/*  Get next symbol  */
/*****************************************************************************/

static void  next_symbol ()
{
  if (Source == NULL) {
    MonetSysError ("Source is null");
    return;
  }
  Current = Source;
  Source = Source->next;
  Current->next = NULL;
  if (InExpression) {
    switch (Current->type) {
    case lex_pipe:
      MonetTree_change_type (Current, lex_bitwise_or);
      break;
    case lex_pipe_out:
      MonetTree_change_type (Current, lex_greater_less);
      break;
    case lex_pipe_in:
      MonetTree_change_type (Current, lex_greater_less);
      break;
    case lex_pipe_append_out:
      MonetTree_change_type (Current, lex_shift_right);
      break;
    case lex_pipe_append_in:
      MonetTree_change_type (Current, lex_shift_left);
      break;
    default:
      break;
    }
  }
}

/*****************************************************************************/
/*  Put node onto debris list  */
/*****************************************************************************/

static void DEBRIS(node)
     MonetTree node;
{
#ifdef MonetDebug
  if (node -> debris) {
    MonetSysError ("Debris [%s][%d] is already debris !!!",
	       LNAME(node), node->type);
  }
  if (node -> arg) {
    MonetSysError ("Debris [%s][%d] has argument !!!", LNAME(node), node->type);
  }
  if (node -> next) {
    MonetSysError ("Debris [%s][%d] has a next node !!!", LNAME(node), node->type);
  }
  switch (node->type) {
  case lex_end:
  case lex_sep:
  case lex_end_braces:
  case lex_end_array:
  case lex_end_parenthesis:
    break;
  default:
    MonetSysError ("Debris [%s][%d] this type should never be debris !!!",
	       LNAME(node), node->type);
    break;
  }
  node -> debris = True;
#endif
  node -> next = Debris;
  Debris = node;
}


#define LOOK_AHEAD \
     (Source->type)



/***************************************************************************/
/*  Parse braces */
/***************************************************************************/

static MonetTree parse_braces ()
{
  MonetTree tree;

  tree = Current;
  tree->arg = parse_statement ();
  if (tree -> arg) {
    tree->arg->argc = 1;
  }
  if (!Error) {
    if (Current->type == lex_end_braces) {
      /* Must free Current because '}' is not used */
      DEBRIS(Current);
      switch (LOOK_AHEAD) {
	/*  insert a seperator in the lex statement  */
      case lex_call_function:
      case lex_parenthesis:
      case lex_braces:
      case lex_var:
      case lex_number:
      case lex_string:
	Current = MonetTree_create (lex_sep, ";");
	break;
      default:
	next_symbol ();
	break;
      }
    } else if (!Error) {
      MonetParseError ("Missing '}', found [%s]", LNAME(Current));
      Error = True;
    }
  }
  return tree;
}

/***************************************************************************/
/*  Parse parenthesis brackets */
/***************************************************************************/

static MonetTree parse_para ()
{
  MonetTree tree;

  tree = Current;
  InExpression = True;
  tree->arg = parse_statement ();
  if (!Error) {
    if (Current->type == lex_end_parenthesis) {
      DEBRIS(Current);
      if (tree -> arg) {
	tree->arg->argc = 1;
      }
    } else {
      MonetParseError ("Missing ')', found [%s]", LNAME(Current));
      Error = True;
    }
  }
  InExpression = False;
  return tree;
}

/***************************************************************************/
/*  Parse array [] brackets */
/***************************************************************************/

static MonetTree parse_array ()
{
  MonetTree tree;

  InExpression = True;
  tree = parse_statement ();
  if (!Error) {
    if (Current->type == lex_end_array) {
      DEBRIS(Current);
    } else {
      MonetParseError ("Missing ']', found [%s]", LNAME(Current));
      Error = True;
    }
  }
  InExpression = False;
  return tree;
}

/***************************************************************************/
/*  Parse postfix operators  */
/***************************************************************************/

static MonetTree parse_postfix (tree)
     MonetTree tree;
{
  switch (Current->type) {
  case lex_question:
  case lex_inc_dec:
    Current -> arg = tree;
    tree = Current;
    tree -> arg -> argc = 1;
    next_symbol ();
    break;
  case lex_open_array:
    Current -> arg = tree;
    tree = Current;
    if (tree -> arg -> next = parse_array ())
      tree -> arg -> argc = 2;
    else
      tree -> arg -> argc = 1;
    if (! Error)
      next_symbol ();
    break;
  default:
    break;
  }
  return tree;
}

/***************************************************************************/
/*  Parse terms */
/***************************************************************************/

static MonetTree parse_simple_term ()
{
  MonetTree tree;

  next_symbol ();

  switch (Current->type) {
  case lex_parenthesis:
    tree = parse_para ();
    break;
  case lex_string:
  case lex_number:
    tree = Current;
    break;
  case lex_var:
    tree = Current;
    if ((tree->arg = parse_arg ()) && !Error) {
      tree -> arg -> argc = 1;
    } else {
      Error = True;
      MonetParseError ("Arg expected after [%s], [%s] found",
		       LNAME(tree), LNAME(Current));
    }
    break;
  default:
    tree = NULL;
    break;
  }
  return tree;
}

/***************************************************************************/
/*  Parse a single argument  */
/***************************************************************************/

static MonetTree parse_arg ()
{
  MonetTree tree;

  if (tree=parse_simple_term ()) {
  } else if (Error) {
  } else if (Current->type == lex_call_function) {
    tree = Current;
    MonetTree_change_type (tree, lex_string);
  }
  if (tree && ! Error) {
    tree = parse_postfix (tree);
  }
  return tree;
}

/***************************************************************************/
/*  Parse function call  */
/***************************************************************************/

static MonetTree parse_call ()
{
  MonetTree tree, list;

  tree = Current;
  if (list = parse_arg ()) {
    tree->arg = list;
    while (! Error) {
      ++(tree->arg->argc);
      if (list->next = parse_arg ()) {
	list = list->next;
      } else {
	break;
      }
    }
    if (Current->type == lex_braces && ! Error) {
      list->next = parse_braces ();
      ++(tree->arg->argc);
    }
  } else {
    switch (Current->type) {
    case lex_braces:
      if (tree->arg = parse_braces ())
	tree->arg->argc = 1;
    default:
      break;
    }
  }
  return tree;
}

/***************************************************************************/
/*  Parse term  */
/***************************************************************************/

static MonetTree parse_term (constant)
     MonetBool constant;
{
  MonetTree tree;

  if ((tree=parse_simple_term ()) && !Error) {
    next_symbol ();
  } else if (! Error) {
    switch (Current->type) {
    case lex_call_function:
      if (constant) {
	tree = Current;
	MonetTree_change_type (tree, lex_string);
	next_symbol ();
      } else {
	tree = parse_call ();
      }
      break;
    case lex_braces:
      tree = parse_braces ();
      break;
    case lex_end_braces:
    case lex_end_parenthesis:
    case lex_end_array:
      break;
    default:
      Error = True;
      MonetParseError ("Simple term expected, [%s] found", LNAME(Current));
    }
  }
  if (tree && ! Error) {
    tree = parse_postfix (tree);
  }
  return tree;
}

/*****************************************************************************/
/*  Parse expressions and operators */
/*****************************************************************************/

static MonetTree parse_entity_name ()
{
  MonetTree tree;
  if ((tree = parse_term (False)) && !Error) {
    switch (Current->type) {
    case lex_entity_name:
    case lex_dot:
    case lex_get_attribute:
      if (tree -> type == lex_call_function) {
	tree -> type = lex_string;
      }
      Current->arg = tree;
      tree = Current;
      if (tree->arg->next = parse_entity_name ()) {
	tree->arg->argc = 2;
      } else if (! Error) {
	Error = True;
	MonetParseError ("Expression expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_unary ()
{
  MonetTree tree;

  switch (LOOK_AHEAD) {
  case lex_sub:
  case lex_add:
  case lex_not:
  case lex_bitwise_not:
  case lex_unary_minus:
  case lex_unary_plus:
  case lex_inc_dec:
    next_symbol ();
    if (Current->type == lex_add) {
      MonetTree_change_type (Current, lex_unary_plus);
      MonetTree_change_name (Current, "unary+");
    } else if (Current->type == lex_sub) {
      MonetTree_change_type (Current, lex_unary_minus);
      MonetTree_change_name (Current, "unary-");
    }
    tree = Current;
    if ((tree->arg = parse_entity_name ()) && !Error) {
      tree -> arg -> argc = 1;
    } else {
      Error = True;
      MonetParseError ("Entity name expression expected after [%s],[%s] found",
		       LNAME(tree), LNAME(Current));
    }
    break;
  default:
    if ((tree = parse_entity_name ()) && !Error) {
    }
    break;
  }
  return tree;
}

static MonetTree parse_mul_div_expression ()
{
  MonetTree tree;
  if ((tree = parse_unary ()) && !Error) {
    switch (Current->type) {
    case lex_mul_div:
      Current->arg = tree;
      tree = Current;
      if (tree->arg->next = parse_mul_div_expression ()) {
	tree->arg->argc = 2;
      } else {
	MonetParseError ("Expression expected after [%s]", LNAME(tree));
	Error = True;
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_simple_expression ()
{
  MonetTree tree;
  if ((tree = parse_mul_div_expression ()) && !Error) {
    switch (Current->type) {
    case lex_add:
    case lex_sub:
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_simple_expression ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	Error = True;
	MonetParseError ("Expression expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_shift ()
{
  MonetTree tree;
  if ((tree = parse_simple_expression ()) && !Error) {
    switch (Current->type) {
    case lex_shift_left:
    case lex_shift_right:
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_shift ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	Error = True;
	MonetParseError ("Shift Bits expression expected after [%s]",
			 LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_greater_less ()
{
  MonetTree tree;
  if ((tree = parse_shift ()) && !Error) {
    switch (Current->type) {
    case lex_greater_less:
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_greater_less ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	Error = True;
	MonetParseError ("Simple expression  expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_eq_ne ()
{
  MonetTree tree;
  if ((tree = parse_greater_less ()) && !Error) {
    switch (Current->type) {
    case lex_eq_ne:
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_eq_ne ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	Error = True;
	MonetParseError ("Expression expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_bitwise ()
{
  MonetTree tree;
  if ((tree = parse_eq_ne ()) && !Error) {
    switch (Current->type) {
    case lex_bitwise_or:
    case lex_bitwise_and:
    case lex_bitwise_xor:
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_bitwise ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	Error = True;
	MonetParseError ("Bitwise expression expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_assign ()
{
  MonetTree tree;
  if ((tree = parse_bitwise ()) && !Error) {
    switch (Current->type) {
    case lex_assign:
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_assign ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	Error = True;
	MonetParseError ("Expression expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_and_or ()
{
  MonetTree tree;
  if ((tree = parse_assign ()) && !Error) {
    if (tree->type == lex_and || tree->type == lex_or) {
      MonetParseError ("Statement expected before [%s]", LNAME(tree));
    } else {
      switch (Current->type) {
      case lex_and:
      case lex_or:
	Current->arg = tree;
	tree = Current;
	if ((tree->arg->next = parse_and_or ()) &&
	    tree->type != lex_end && !Error) {
	  tree->arg->argc = 2;
	} else {
	  Error = True;
	  MonetParseError ("Statement expected after [%s]", LNAME(tree));
	}
	break;
      default:
	break;
      }
    }
  }
  return tree;
}

static MonetTree parse_pipe ()
{
  MonetTree tree;
  if ((tree = parse_and_or ()) && !Error) {
    switch (Current->type) {
    case lex_pipe:
      Current->arg = tree;
      MonetTree_change_name (Current, "pipe|");
      tree = Current;
      if ((tree->arg->next = parse_pipe ()) && !Error) {
	tree->arg->argc = 2;
      } else {
	MonetParseError ("Instruction expected after [%s]", LNAME(tree));
	Error = True;
      }
      break;
    case lex_pipe_in:
    case lex_pipe_out:
    case lex_pipe_append_out:
    case lex_pipe_append_in:
      {
	char buf [10];
	(void) sprintf (buf, "pipe%s", LNAME(Current));
	MonetTree_change_name (Current, buf);
      }
      Current->arg = tree;
      tree = Current;
      if ((tree->arg->next = parse_term (True)) && !Error) {
	tree->arg->argc = 2;
      } else if (! Error) {
	Error = True;
	MonetParseError ("Conditional expected after [%s]", LNAME(tree));
      }
      break;
    default:
      break;
    }
  }
  return tree;
}

static MonetTree parse_statement ()
{
  MonetTree tree;

  switch (LOOK_AHEAD) {
  case lex_sep:
    next_symbol ();
    DEBRIS(Current);
    Current = NULL;
    break;
  case lex_end:
    return NULL;
  default:
    break;
  }
  while ((tree = parse_pipe ()) && tree->type == lex_sep && !Error);
  if (tree && ! Error)
    switch (Current->type) {
    case lex_sep:
      Current->arg = tree;
      tree = Current;
      if (tree->arg->next = parse_statement ()) {
      }
      break;
    default:
      break;
    }
  return tree;
}

/***************************************************************************/
/*  Top level parsing routine  */
/***************************************************************************/

extern MonetTree MonetTree_parse (tree)
     MonetTree tree;
{
  if (tree) {
    Error = False;
    Source = tree;
    Current = NULL;
    Debris = NULL;

    InExpression = False;

    if (Monet_debug() > 99 || Monet_debug() == 5) {
      MONETDEBUG(5, ( "Parse [%s]", MonetTree2String (tree)));
    }
    if (tree = parse_statement ()) {
      if (Current && ! Error) {
	switch (Current->type) {
	case lex_end:
	  break;
	default:
	  Error = True;
	  MonetParseError ("Statement expected after [%s], not [%s]",
			 LNAME(tree), LNAME(Current));
	  break;
	}
      } else if (Monet_debug() > 99 || Monet_debug() == 5) {
	MONETDEBUG(5, ( "Parsed [%s]", MonetTree2String (tree)));
	if (Monet_debug() > 99) {
	  MonetTree_dump (tree, (unsigned) 0);
	  MonetTree_dump (Debris, (unsigned) 0);
	}
      }
    }
    if (Error) {
      MonetTree_free (tree);
      MONETDEBUG(1, ("Parse failed"));
      tree = NULL;
    }
    if (Current)
      MonetTree_free (Current);
    MonetTree_free (Debris);
    MonetTree_free (Source);
  }
  return tree;
}

/*****************************************************************************/
/*****************************************************************************/
/* $Header: script_parse.c,v 1.1 91/09/20 10:11:59 a38 Exp $ */
/*****************************************************************************/
/*****************************************************************************/
