/********************************************************************/
/********************************************************************/
/* $Header$ */
/********************************************************************/
/********************************************************************/

#include <monet/msh/General.h>
#include <monet/msh/Script.h>
#include "x25P.h"

#include <ctype.h>
#include <sys/stat.h>

static String x121_file;

/**********************************************************************/
/*  Definitions  */
/**********************************************************************/

static MonetLookupX25 Lookup;

static int LookupEntries = 0;
static time_t FileLoadedTime = 0;
static long X25_delay;

static void clear_addresses ()
{
  MonetLookupX25 l;
  FORALL (l, Lookup, next) {
    MonetFree (l->host);
    MonetFree (l->address);
    MonetFree (l->userdata);
    MonetFree(l);
  }
  Lookup = NULL;
}

/*********************************************************************/
/*  Read in a user data string definition and prodcue a byte string.  */
/**********************************************************************/

static void process_string (data, len)
     char * data;
     int *len;
{
  char special;
  char * c, *r;

  r = c = data;
  for ((*len)=0; *c!=NULL;c++, r++, (*len)++) {
    if (*len == 1023) {
      MonetPrintfError ("Length of user data is greater than [%d]", 1024);
      break;
    }
    switch (*c) {
    case '\\':
      c++;
      switch (*c) {
      case 'n': *r = '\n'; break;
      case 'r': *r = '\r'; break;
      default:
	if (isdigit(*c)) {
	  for (special= *c-'0',c++;isdigit(*c);c++)
	    special = special * 10 + *c - '0';
	  *r = special;
	} else
	  *r = *c;
	break;
      }
      /*  Ignore double quotes, this is the NULL character  */
    case '\"': c++; break;
    default:
      *r = *c;
      break;
    }
  }
  *r = NULL;
}

/***************************************************************************/

/***************************************************************************/

static MonetLookupX25 create (host, address, udata)
     String host, address, udata;
{
  MonetLookupX25 lookup;

  char userdata[1024]; /* !!! */
  lookup = MonetNew (struct monet_lookup_x25);
  lookup->host    = MonetNewString (host);
  lookup->address = MonetNewString (address);
  lookup->addrlen = MonetStrlen (address);
  lookup->next = Lookup;
  Lookup = lookup;
  if (udata == NULL) {
    strcpy (userdata, "\\1\\0\\0\\0");
  } else {
    strcpy (userdata, udata);
  }
  process_string (userdata, &(lookup->userdatalen));
  lookup->userdata = MonetMalloc (lookup->userdatalen);
  bcopy (userdata, lookup->userdata, lookup->userdatalen);
  if (udata) {
    MONETDEBUG(10, ("[%s] user data len %d, data %x [%s]",
		host,
		lookup->userdatalen,
		(int) *(lookup->userdata), lookup->userdata));
  }
  LookupEntries ++;
  return lookup;
}

/***************************************************************************/

/***************************************************************************/

static MonetBool load_file (name)
     char * name;
{
  clear_addresses ();
  Script_source_file (name, ".msh", "Cannot open X.25 host/address file");
  MONETDEBUG(10, ("%d entries in the x25 translation table", LookupEntries));
  FileLoadedTime = time (0);
  return True;
}

extern MonetLookupX25 MonetHost2X25 (host)
     char * host;
{
  MonetLookupX25 l;

  FORALL (l, Lookup, next)
    if (MonetStringsEqual(l->host, host))
      return l;

  return NULL;
}

static void check_x25_file ()
{
  struct stat stats;
  char * name;
  
  if (name = (char *) MonetFilename (x121_file)) {
    if (stat (name, &stats) != 0) {
      MONETDEBUG(10, ("Error doing stat on [%s]", name));
      return;
    }
    if (stats.st_mtime > FileLoadedTime) {
      MONETDEBUG(10, ("X25 file modified, reloading [%s]", name));
      /*  Should free previous entries first  */
      clear_addresses ();
      load_file (x121_file);
    }
  }
}

/*********************************************************************/
/*  Check if file was changed  */
/*********************************************************************/

extern void MonetX25Lookup_periodic_check_file ()
{
  static long    checked_x25_file = 0;

  if (checked_x25_file + X25_delay < time (NULL)) {
    check_x25_file ();
  }
}

/*****************************************************************************/
/*  X.25  Load file */
/*****************************************************************************/


MONETCALLBACK(do_x25_address)
     if (argc == 3) {
       (void) create (argv[1], argv[2], (char *) NULL);
     } else {
       (void) create (argv[1], argv[2], argv[3]);
     }
MONETEND

MONETCALLBACK(do_clear_x25_addresses)
     clear_addresses ();
MONETEND

MONETCALLBACK(do_help_x25_address)
     MonetLookupX25 l;

     FORALL (l, Lookup, next) {
       MonetPrintfnl ("[%s] [%s] [%s]", l->host, l->address, l->userdata);
     }
     MonetPrintPrettyCount (NULL, "X.25 host/address", LookupEntries);
MONETEND

MONETCALLBACK(do_x25_host_to_address)
     MonetLookupX25 l;
     l = MonetHost2X25 (argv[1]);
     result = l->address;
MONETEND

MONETCALLBACK(do_check_x25_file)
     check_x25_file ();
MONETEND

MONETCALLBACK(do_load_x25_file)
     load_file (x121_file);
MONETEND

/********************************************************************/
/*  Init and finish  */
/********************************************************************/

extern void MonetX25Lookup_finish ()
{
  
}

extern void MonetX25Lookup_init ()
{
  FileLoadedTime = 0;

  X25_delay = 660;

  (void) MonetMember_module_class ("x25");
  MonetMember_create
    ("host-to-address",  do_x25_host_to_address,
     "hostname",
     "Returns x25 address of host");
  MonetMember_create
    ("address",   do_x25_address,
     "host address ?user-data",
     "Adds an X.25 address to the translation table. \
The user data is used when 'pinging' the host.");
  MonetMember_create
    ("clear-addresses",  do_clear_x25_addresses,
     "",
     "Removes all x25 host/address translations");
  MonetMember_create
    ("help-address",  do_help_x25_address,
     "",
     "Lists all defined x25 addresses");
  MonetMember_create
    ("load-file",  do_load_x25_file,
     "",
     "Loads/reloads the default x25 file. \
Uses the configuration parameter 'config x121_file'");
  MonetMember_create
    ("check-file",  do_check_x25_file,
     "",
     "Checks to see if x25 file has been changed and if so reloads it. \
Uses the configuration parameter 'config x121_file'");
  MonetMember_create_write ( "file", &x121_file,
			    "File name of x25 host/address translation table");
  MonetMember_create_write_long ( "file-delay", &X25_delay,
				 "Delay between checking x25 file changed");
  MonetConfig_create ("-x25",  "x25->file", NULL,
		      "The name of the x121 host/address table");
  MonetConfig_create ( "-x25delay", "x25->file-delay", NULL,
		      "Period between checking if X.25 host file has changed");
}


/********************************************************************/
/********************************************************************/
/* $Header$ */
/********************************************************************/
/********************************************************************/


